"""
FileManager manages current directory and formatter.
"""
import os
import time
import subprocess
from gcoreutils.io_util import read_dict, write_dict
from gcoreutils.latex_util import check_latex_installed
from multipie import get_binary
from multipie.group.point_group import PointGroup
from multipie.group.space_group import SpaceGroup


# ==================================================
class MultiPieManager:
    """
    manage directory, read, write etc.
    """

    # ==================================================
    def __init__(self, topdir=None, verbose=False, symbolic=True, formatter=True, pdf=True, parallel=True, qtdraw=True):
        """
        initialize the class.

        Args:
            topdir (str): top directory for output, ends without slash.
            verbose (bool, optional): verbose parallel info. ?
            symbolic (bool, optional): output in sympy format ?
            formatter (bool, optional): format by using black ?
            pdf (bool, optional): create pdf/tex file ?
            parallel (bool, optional): use parallel code ?
            qtdraw (bool, optional): use QtDraw ?
        """
        self._start = time.time()
        self._lap = self._start

        # initialize.
        if topdir is not None:
            os.makedirs(os.path.abspath(topdir), exist_ok=True)
            os.chdir(topdir)

        self._topdir = os.getcwd().replace(os.sep, "/")
        self._dirname = self._topdir

        self._core = get_binary()

        self._verbose = verbose
        self._symbolic = symbolic
        self._formatter = formatter
        self._pdf = pdf
        if not check_latex_installed():
            self._pdf = False
        self._parallel = parallel
        self._qtdraw = qtdraw
        try:
            import qtdraw
        except ImportError:
            self._qtdraw = False

        self._group = PointGroup("C1", core=self._core)
        self._point_group = self._group
        self._molecule = True

    # ==================================================
    @property
    def dirname(self):
        return self._dirname

    # ==================================================
    @property
    def core(self):
        return self._core

    # ==================================================
    @property
    def verbose(self):
        return self._verbose

    # ==================================================
    @property
    def symbolic(self):
        return self._symbolic

    # ==================================================
    @property
    def pdf(self):
        return self._pdf

    # ==================================================
    @property
    def parallel(self):
        return self._parallel

    # ==================================================
    @property
    def qtdraw(self):
        return self._qtdraw

    # ==================================================
    @property
    def group(self):
        return self._group

    # ==================================================
    @property
    def point_group(self):
        return self._point_group

    # ==================================================
    @property
    def molecule(self):
        return self._molecule

    # ==================================================
    def elapsed(self, from_stamp=True):
        """
        elapsed time.

        Args:
            from_stamp (bool, optional): if True, return difference from lap.

        Returns:
            float: elapsed time.
        """
        if from_stamp:
            return time.time() - self._lap
        else:
            return time.time() - self._start

    # ==================================================
    def set_group(self, group):
        """
        set group (and point group).

        Args:
            group (str or int or TagGroup): group.

        Notes:
            - space/point group can be accessed by self.group property.
            - associated point group can be accessed by self.point_group property.
        """
        if type(group) not in [str, int]:
            group = str(group)
        self._molecule = not (type(group) == int or "^" in group)
        if self._molecule:
            self._group = PointGroup(group, core=self._core)
            self._point_group = self._group
        else:
            self._group = SpaceGroup(group, core=self._core)
            self._point_group = self._group.pg

    # ==================================================
    def set_stamp(self):
        """
        set current time.
        """
        self._lap = time.time()

    # ==================================================
    def log(self, text, stamp="", end=None, file=None):
        """
        write log if verbose is True.

        Args:
            text (str): text to write.
            stamp (str or None, optional): attach elapsed time if "start" or "", otherwise no stamp is attached.
            end (str, optional): same as print end option.
            file (str, optional): same as print file option.
        """
        if self._verbose:
            if stamp is not None:
                text += " ( " + str(round(self.elapsed(stamp != "start"), 3)) + " [sec] )."
            print(text, end=end, file=file)

    # ==================================================
    def filename(self, filename, full=True):
        """
        get (full) file name.

        Args:
            filename (str): file name.
            full (bool, optional): with full path ?

        Returns:
            str: full file name (with full path).
        """
        if full:
            return self.dirname + "/" + filename
        else:
            return os.path.split(filename)[1]

    # ==================================================
    def read(self, file_dict):
        """
        read dict file or dict itself.

        Args:
            file_dict (str or dict): filename of dict. or dict.

        Returns:
            dict: read dict.
        """
        if type(file_dict) == str:
            full = self._topdir + "/" + file_dict
            if os.path.isfile(full):
                dic = read_dict(full)
                self.log(f"  * read '{full}'.", None)
            else:
                raise Exception(f"cannot open {full}.")
        else:
            dic = file_dict

        return dic

    # ==================================================
    def write(self, filename, dic, header=None, var=None):
        """
        write dict to file.

        Args:
            filename (str): file name.
            dic (dict): dict to write.
            header (str, optional): header of dict.
            var (str, optional): variable name for dict.
        """
        full = self.filename(filename)
        write_dict(full, dic, header, var)
        self.log(f"  * wrote '{filename}'.", None)

    # ==================================================
    def create_subdir(self, subdir):
        """
        create sub directory under topdir.
        """
        self._dirname = os.path.abspath(self._topdir + "/" + subdir)
        os.makedirs(self._dirname, exist_ok=True)

    # ==================================================
    def formatter(self):
        if self._formatter:
            cmd = "black --line-length=130 ."
            if self._qtdraw:
                cmd += " *.qtdw"
            try:
                subprocess.run(cmd, capture_output=True, check=True, cwd=self._dirname, shell=True)
            except subprocess.CalledProcessError:
                raise Exception("Formatting by black is failed.")



import os
import sys

__version__ = "1.1.5"
__top_dir__ = os.path.normpath(os.path.dirname(__file__) + "/../") + "/"
__bin_dir__ = __top_dir__ + "multipie/binary_data/"


# ==================================================
from gcoreutils.string_util import class_name
from gcoreutils.binary_manager import BinaryManager

from multipie.character.character_pg_set import CharacterPGSet
from multipie.harmonics.harmonics_pg_set import HarmonicsPGSet
from multipie.wyckoff.wyckoff_g_set import WyckoffGSet
from multipie.symmetry_operation.symmetry_operation_g_set import SymmetryOperationGSet
from multipie.virtual_cluster.virtual_cluster_pg_set import VirtualClusterPGSet
from multipie.clebsch_gordan.clebsch_gordan_pg_set import ClebschGordanPGSet
from multipie.response_tensor.response_tensor_pg_set import ResponseTensorPGSet
from multipie.multipole.base.base_atomic_multipole_dataset import BaseAtomicMultipoleDataset


# ==================================================
def create_binary(clean=False, verbose=True):
    """
    create binary data.

    Args:
        clean (bool, optional): remove all binaries at beginning.
        verbose (bool, optional): verbose progress ?
    """
    dataset = [
        CharacterPGSet,
        HarmonicsPGSet,
        WyckoffGSet,
        SymmetryOperationGSet,
        VirtualClusterPGSet,
        ClebschGordanPGSet,
        ResponseTensorPGSet,
        BaseAtomicMultipoleDataset,
    ]

    def dprint(s, verbose):
        if verbose:
            print(s, file=sys.stderr)

    if clean:
        for ds in dataset:
            name = class_name(ds) + ".pkl"
            f = __bin_dir__ + name
            if os.path.exists(f):
                os.remove(f)
                dprint("removed " + f, verbose)

    dprint("creating binaries ...", verbose)
    core = BinaryManager(__bin_dir__, verbose=True)
    for ds in dataset:
        core[ds]


# ==================================================
def get_binary(load=False, verbose=False):
    """
    get binary data.

    Args:
        load (bool, optional): load all binaries ?
        verbose (bool, optional): verbose progress ?

    Returns:
        BinaryManager: binary manager.
    """
    core = BinaryManager(__bin_dir__, verbose=verbose)
    if load:
        core.load()
    return core



"""
BaseAtomicMultipoleSet manages a set of atomic multipoles for lm, jm, cubic, and hexagonal basis.
"""
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_multipole import TagMultipole
from multipie.tag.tag_list import TagList
from multipie.data.data_atomic_multipoles import (
    _data_atomic_multipoles_lm,
    _data_atomic_multipoles_lm,
    _data_atomic_multipoles_cubic,
    _data_atomic_multipoles_hexagonal,
)


_data_atomic_multipoles_set = {
    "lm": _data_atomic_multipoles_lm,
    "jm": _data_atomic_multipoles_lm,
    "cubic": _data_atomic_multipoles_cubic,
    "hexagonal": _data_atomic_multipoles_hexagonal,
}


# ==================================================
class BaseAtomicMultipoleSet(dict):  # dict of (multipole tag, matrix) { TagMultipole: Matrix }.
    """
    a set of atomic multipoles for lm, jm, cubic, and hexagonal basis.

    Attributes:
        tag (str): tag
    """

    # ==================================================
    def __init__(self, b_type):
        """
        initialize the class.

        Args:
            b_type (str): basis type, (lm/jm/cubic/hexagonal).
        """
        self.tag = b_type
        """basis type, (lm/jm/cubic/hexagonal)."""

        for mtag, m in _data_atomic_multipoles_set[b_type].items():
            self[TagMultipole(mtag)] = NSArray(m, style="matrix", fmt="sympy", real=False)

    # ==================================================
    def __str__(self):
        return self.tag

    # ==================================================
    def __repr__(self):
        return self.tag

    # ==================================================
    def latex(self):
        return self.tag

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagMultipole(tag))
        else:
            return self.get(tag)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())



"""
create model view, info, and basis.
"""
from multipie.model.material_model import MaterialModel
from multipie.model.multipie_manager import MultiPieManager
from multipie.model.symmetry_adapted_model import SymmetryAdaptedModel
from multipie.model.util.create_pdf import ModelPDF


# ==================================================
def _create_single_model(model_dict, mpm, view_mode):
    """
    create a model view, info, and basis.

    Args:
        model_dict (dict or str): minimal model information.
        mpm (MultiPieManager): multipie manager.
        view_mode (str): mode for QtDraw.
    """
    model_dict = mpm.read(model_dict)
    model_dict = MaterialModel.regularize(model_dict)
    model_name = model_dict["model"]
    mpm.set_group(model_dict["group"])
    mpm.create_subdir(model_dict["option"]["output"])

    mpm.log(f"=== creating files for '{model_name}' in {mpm.dirname} ...", None)

    # create model.
    mpm.log("creating model ... ", None)
    cmodel = MaterialModel(model_dict, mpm)
    mpm.write(model_name + "_model.py", cmodel, cmodel._header(), model_name)

    # create view.
    if mpm.qtdraw:
        mpm.log("creating view ... ", None)
        view = cmodel.create_view(mode=view_mode)
        mpm.write(model_name + "_view.qtdw", view, "\nQtDraw data file in Python dict format.\n")

    # create SAMB.
    mpm.log("creating SAMB ... ", None)
    irrep = cmodel["info"]["generate"]["irrep"]
    if cmodel["info"]["generate"]["time_reversal_type"] == "electric":
        head = ["Q", "G"]
    elif cmodel["info"]["generate"]["time_reversal_type"] == "magnetic":
        head = ["T", "M"]
    else:
        head = ["Q", "G", "T", "M"]
    samb = SymmetryAdaptedModel(cmodel, mpm=mpm, head=head, irrep=irrep, toroidal_priority=False)
    samb_dict = samb.create_dict()
    mpm.write(model_name + "_samb.py", samb_dict, SymmetryAdaptedModel._header(), model_name)

    # create matrix.
    mpm.log("creating SAMB matrix ... ", None)
    samb_matrix = samb.create_matrix(full=True, fmt="sympy")
    mpm.write(model_name + "_matrix.py", samb_matrix, SymmetryAdaptedModel._matrix_header(), model_name)

    # create LaTeX and PDF.
    if mpm.pdf:
        mpm.log("creating LaTeX and PDF ... ", None)
        ModelPDF(cmodel, samb_dict, mpm)

    mpm.formatter()
    mpm.log("=== total elapsed_time:", "start")


# ==================================================
def create_model(
    model_list, topdir=None, symbolic=True, parallel=True, verbose=False, pdf=True, formatter=True, view_mode=None, qtdraw=True
):
    """
    create model view, info, and basis.

    Args:
        model_list (dict or list): minimal model information.
        topdir (str, optional): top directory for output.
        symbolic (bool, optional): output in sympy format ?
        parallel (bool, optional): use parallel code.
        verbose (bool, optional): verbose parallel info.
        pdf (bool, optional): create pdf/tex file.
        formatter (bool, optional): format by using black.
        view_mode (str, optional): mode for QtDraw, standard/arrow/debug. if None, option in the model is used.
        qtdraw (bool, optional): use QtDraw ?
    """
    mpm = MultiPieManager(topdir, verbose, symbolic, formatter, pdf, parallel, qtdraw)

    if type(model_list) == list:
        for model in model_list:
            _create_single_model(model, mpm, view_mode)
    else:
        _create_single_model(model_list, mpm, view_mode)



"""
create SAMB from model input.
"""
import os
import click
from multipie.model.create_model import create_model
from multipie.model.material_model import input_str


# ================================================== create_samb
@click.command()
@click.option("-p", "--parallel", is_flag=True, help="parallel off.")
@click.option("-v", "--verbose", is_flag=True, help="verbose off.")
@click.option("-l", "--latex", is_flag=True, help="LaTeX and PDF off.")
@click.option("-f", "--formatter", is_flag=True, help="formatter off.")
@click.option("-m", "--mode", type=click.Choice(["arrow", "debug"]), default=None, help="view mode.")
@click.option("-d", "--qtdraw", is_flag=True, help="QtDraw off.")
@click.option("-o", "--output", default=".", help="output top-directory name.")
@click.option("-i", "--input", is_flag=True, help="show input format, and exit.")
@click.argument("models", nargs=-1)
def cmd(models, parallel, verbose, latex, formatter, mode, qtdraw, output, input):
    """
    create SAMB from input files.

        MODELS : file names (w or w/o `.py`).
    """
    if input:
        click.echo(input_str)
        exit()
    if len(models) < 1:
        exit()

    rel = os.path.relpath(".", "./" + output)

    models = [i[:-3] if i[-3:] == ".py" else i for i in models]
    models = [os.path.join(rel, i + ".py") for i in models]
    create_model(
        models,
        topdir=output,
        symbolic=True,
        parallel=not parallel,
        verbose=not verbose,
        pdf=not latex,
        formatter=not formatter,
        view_mode=mode,
        qtdraw=not qtdraw,
    )



"""
SymmetryOperation manages symmetry operation.
"""
from gcoreutils.nsarray import NSArray
from multipie.symmetry_operation.util.symmetry_operation_util import (
    to_cartesian,
    to_reduced,
    to_axial,
    reflection_matrix,
    rotation_matrix,
    rotoinversion_matrix,
)
from multipie.tag.tag_symmetry_operation import TagSymmetryOperation


# ==================================================
class SymmetryOperation:
    """
    symmetry operation.

    Attributes:
        tag (TagSymmetryOperation): symmetry-operation tag.
        m_polar (NSArray): symmetry-operation matices for polar vector, (3x3:point group) or (4x4:space group).
        m_axial (NSArray): symmetry-operation matices for axial vector, (3x3:point group) or (4x4:space group).
    """

    # ==================================================
    def __init__(self, so_tag, crystal):
        """
        initialize the class.

        Args:
            so_tag (TagSymmetryOperation or str): symmetry-operation tag.
            crystal (str): crystal type, (triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic).
        """
        so_tag = TagSymmetryOperation(str(so_tag))
        self.tag = so_tag
        """symmetry-operation tag."""

        so = SymmetryOperation._matrix(self.tag, crystal)
        self.m_polar = so[0]
        """symmetry-operation matices for polar vector, (3x3:point group) or (4x4:space group)."""
        self.m_axial = so[1]
        """symmetry-operation matices for axial vector, (3x3:point group) or (4x4:space group)."""

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    @classmethod
    def _matrix(cls, tag, crystal):
        """
        symmetry operation matrix.

        Args:
            tag (TagSymmetryOperation): tag of symmetry operation.
            crystal (str): crystal.

        Returns:
            - NSArray: symmetry-operation matices for polar vector, (3x3:point group) or (4x4:space group).
            - NSArray: symmetry-operation matices for axial vector, (3x3:point group) or (4x4:space group).
        """
        axis = NSArray(tag.axis)
        axis = to_cartesian(crystal, axis)

        if tag.mirror:
            h = reflection_matrix(axis)
        elif tag.inversion:
            h = rotoinversion_matrix(tag.n, axis)
        else:
            h = rotation_matrix(tag.n, axis)

        ha = to_axial(h)

        h = to_reduced(crystal, h)
        ha = to_reduced(crystal, ha)

        if tag.is_point_group():
            return h, ha
        else:
            t = NSArray(tag.t)
            h4 = h._pad()
            h4[0:3, 3] = t[:]
            ha4 = ha._pad()
            ha4[0:3, 3] = t[:]
            return h4, ha4



"""
MaterialModel manages model information of cluster or crystal system.
"""
import numpy as np
from gcoreutils.nsarray import NSArray
from gcoreutils.crystal_util import cell_transform_matrix
from multipie.multipole.util.atomic_orbital_util import parse_orb_list, sort_orb_list, split_orb_list_rank_block, rank
from multipie import __version__

try:
    from qtdraw.qt_draw import QtDraw
    from qtdraw.multipie.setting import default_style

    _qtdraw_available = True
except ImportError:
    _qtdraw_available = False


_default_max_neighbor = 10
_default_search_cell_range = (-2, 3, -2, 3, -2, 3)


# ==================================================
input_str = """
=== Input for SAMB construction (* optional [default]) ===
- model : model name (str).
- group : group name (Schoenflies notation) or group number (space group) (str/int).
- site : site position, orbital info. (dict) { name: ("position", orbital or orbital list) }.
- bond : bond (list) [ ("tail", "head", (list of) neighbors) ].
- spinful* : spinful basis ? (bool), [False].
- cell* : unit-cell constants (dict) { "a", "b", "c", "alpha", "beta", "gamma" }, [a=b=c=1,alpha=beta=gamma=90].
- option*
  - view* : view point (int list), [None].
  - view_mode* : mode for QtDraw file (str) ("standard"/"arrow"/"debug"), ["standard"].
  - output* : output folder (str), [model name].
  - minimal_samb* (bool) : minimal output in _samb.pdf ? [True].
- generate*
  - model_type* : model type (str), ("tight_binding"/"phonon"), ["tight_binding"].
  - time_reversal_type* : time-reversal type (str), ("electric"/"magnetic"/"both"), ["electric"].
  - irrep* : irrep. (str list), [identity irrep.].
- k_point* : k-point (dict) {name: "position"}, [{ "Γ": "[0,0,0]", "X": "[1/2,0,0]" }].
- k_path* : k-path (str) (concatenate by "-" or "\|"), ["Γ-X"].
"""

# ==================================================
header_str = """
=== Molecule or Crystal Model (* only for crystal) ===
- info
    - model : model name.
    - molecule : molecule or crystal ?
    - group : (tag, detailed str)
    - crystal : crystal class
    - cell* : {a, b, c, alpha, beta, gamma}
    - volume* : unit cell volume
    - a1* : unit cell a1 vector
    - a2* : unit cell a2 vector
    - a3* : unit cell a3 vector
    - option :
        - view : view index
        - view_mode : QtDraw mode, standard/arrow/debug
        - output : output directory.
        - minimal_samb : output minimal SAMB ?
    - generate :
        - model_type : tight_binding/phonon
        - time_reversal_type : electric/magnetic/both
        - irrep : irrep list
    - k_point* : representative k points
    - k_path* : high-symmetry line in k space
    - dimension : dimension of full matrix
    - spinful : spinful or not
    - orbital : list of orbitals (U,D: up/down spin)
    - ket : ket basis list, orbital@site
    - ket_site : list of sites
    - site : input for "site" { name: (position, orbitals) }
    - rep_site : representative site { name: (position, wp, orbitals) }
    - cell_site : { name_idx(pset): (position, SOs) }
    - bond : input for "bond" [ (tail, head, neighbors) ]
    - rep_bond : representative bond { name: (vector@center, wp, directional, neighbor) }
    - cell_bond : { name_idx(pset): (vector@center, SOs) }

- name
    - alias : { cluster_tag: name or name: cluster_tag }
    - site : { site_tag: (name, no) }
    - site_name : { position : (site_tag, pset) }
    - bond : { bond_tag: (tail:head:n:multiplicity, no) }
    - bond_name : { vector@center : (bond_tag, pset) }

- data
    - plus_set* : [ plus_set list ]
    - cluster_site : { cluster_tag: site_list }
    - cluster_bond : { cluster_tag: bond_list }
    - site : { site_tag: (position, SO, (bra_site_no, ket_site_no)) }
    - bond : { bond_tag: (vector@center, SO, (bra_site_no, ket_site_no), vector, tail;head) }
    - cluster_atomic : { (bra_site_no, ket_site_no): [(bra_no, ket_no, matrix_tag)] }
    - atomic_braket : { matrix_tag : (bra_list, ket_list) }

- detail
    - rep_bond_all : { tail_head: [rep_bond in 0-9th neighbors] }
    - cell_range* : search range for bonds
    - max_neighbor : max. of neighbors to search
    - A* : transform matrix, [a1,a2,a3]
    - version : MultiPie version
"""


# ==================================================
class MaterialModel(dict):
    """
    molecule or crystal model information.
    """

    # Attributes:
    #     _mpm (MultiPieManager): MultiPieManager.
    #     _cell (dict): cell info.
    #     _volume (float): cell volume.
    #     _A (NSArray): unit vectors in each column (3x3).
    #     _G (NSArray): metric tensor (3x3).
    #     _A_norm (NSArray): normalized unit vectors.
    # ==================================================
    def __init__(self, dic, mpm):
        """
        initialize the class.

        Args:
            dic (dict): model info. dict.
            mpm (MultiPieManager): MultiPie manager.
        """
        info = {
            "model": "default",
            "molecule": False,
            "group": (1, ""),
            "crystal": "monoclinic",
            "cell": {"a": 1.0, "b": 1.0, "c": 1.0, "alpha": 90.0, "beta": 90.0, "gamma": 90.0},
            "volume": 1.0,
            "a1": "[1.0, 0.0, 0.0]",
            "a2": "[0.0, 1.0, 0.0]",
            "a3": "[0.0, 0.0, 1.0]",
            "option": {"view": [0, 0, 1], "view_mode": "standard", "output": "material_model", "minimal_samb": True},
            "generate": {"model_type": "tight_binding", "time_reversal_type": "electric", "irrep": ["A"]},
            "k_point": {},
            "k_path": "",
            "dimension": 1,
            "spinful": True,
            "orbital": [],
            "ket": [],
            "ket_site": [],
            "site": {},
            "rep_site": {},
            "cell_site": {},
            "bond": [],
            "rep_bond": {},
            "cell_bond": {},
        }
        name = {
            "alias": {},
            "site": {},
            "site_name": {},
            "bond": {},
            "bond_name": {},
        }
        data = {
            "plus_set": [],
            "cluster_site": {},
            "cluster_bond": {},
            "site": {},
            "bond": {},
            "cluster_atomic": {},
            "atomic_braket": {},
        }
        detail = {
            "rep_bond_all": {},
            "cell_range": _default_search_cell_range,
            "max_neighbor": _default_max_neighbor,
            "A": "[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]",
            "version": __version__,
        }

        self._mpm = mpm

        info["molecule"] = self._mpm.molecule

        info["spinful"] = dic["spinful"]
        info["option"] = dic["option"]

        if info["molecule"]:
            del info["cell"], info["volume"], info["a1"], info["a2"], info["a3"], info["k_point"], info["k_path"]
            del data["plus_set"]
            del detail["cell_range"], detail["A"]

        info["model"] = dic["model"]
        info["group"] = (str(self._mpm.group), self._group_str(info["molecule"]))
        info["crystal"] = self._mpm.group.tag.crystal

        cell_info = cell_transform_matrix(cell=dic["cell"], crystal=info["crystal"], translation=False)
        self._cell = cell_info[0]
        self._volume = cell_info[1]
        self._A = cell_info[2]
        self._G = cell_info[3]
        self._A_norm = cell_info[4]
        if not info["molecule"]:
            info["cell"] = self._cell
            info["volume"] = self._volume
            info["a1"] = str(self._A[:, 0])
            info["a2"] = str(self._A[:, 1])
            info["a3"] = str(self._A[:, 2])

        info["generate"] = dic["generate"]
        irrep = info["generate"]["irrep"]
        if irrep is None:
            irrep = self._mpm.point_group.identity_irrep
        if type(irrep) == str:
            irrep = [irrep]
        info["generate"]["irrep"] = irrep

        if not info["molecule"]:
            info["k_point"] = dic["k_point"]
            info["k_path"] = dic["k_path"]
            detail["A"] = str(self.A)
            data["plus_set"] = self._mpm.group.symmetry_operation.plus_set.str()

        if "site" in dic.keys():
            info["site"] = dic["site"]
            self._set_rep_site(info)
            orbital = self._set_site(info, name, data)
            ket_dict = self._set_orbital(info, name, orbital)

        if "bond" in dic.keys():
            info["bond"] = dic["bond"]
            self._set_rep_bond(info, name, detail)
            self._set_bond(info, name, data)

        self._set_matrix(data, orbital, ket_dict)

        self["info"] = info
        self["name"] = name
        self["data"] = data
        self["detail"] = detail

    # ==================================================
    @property
    def model(self):
        return self["info"]["model"]

    # ==================================================
    def create_view(self, scale=True, mode=None):
        """
        create QtDraw file.

        Args:
            scale (bool, optional): objects are scaled by means of volume.
            mode (str, optional): view mode, standard/arrow/debug. if None, option in the model is used.

        Returns:
            dict: QtDraw dict.
        """
        if not _qtdraw_available:
            raise Exception("QtDraw is not installed.")

        info = self["info"]
        name = self["name"]
        data = self["data"]
        molecule = info["molecule"]

        if mode is None:
            mode = info["option"]["view_mode"]

        if molecule:
            n_pset = 1
        else:
            n_pset = len(data["plus_set"])

        qtdraw = QtDraw(
            model=info["model"],
            cell=self._cell,
            view=info["option"]["view"],
            axis_type="abc",
            cluster=molecule,
            clip=False,
        )
        qtdraw.set_crystal(info["crystal"])
        if scale:
            scale = self._volume ** (1 / 3)
        else:
            scale = 1.0

        cluster_site_n = len(default_style["site"])
        cluster_bond_n = len(default_style["bond"])

        # plot sites.
        for pos, (site_name, pset) in name["site_name"].items():
            head, idx = name["site"][site_name]
            cluster_no = int(name["alias"][head].split("_")[1])
            prop_no = cluster_no - 1 if cluster_no < cluster_site_n else cluster_site_n - 1
            color, size, opacity = default_style["site"][prop_no]
            if n_pset > 1:
                cluster_name = f"S{cluster_no}({pset})"
            else:
                cluster_name = f"S{cluster_no}"
            mp = data["site"][site_name][1]
            if mode == "debug":
                site_no = int(site_name.split("_")[1])
                label = f"s{site_no}"
            else:
                label = f"{head}_{idx}: " + self._mapping_str(mp)
            qtdraw.plot_site(pos, name=cluster_name, label=label, color=color, size=size * scale, opacity=opacity)
            if mode != "standard" and idx == 1 and pset == 1:
                qtdraw.plot_site(pos, name="Z", label=label, color="yellow", size=size * scale * 1.2, opacity=0.3)

        # plot bonds.
        for bond, (bond_name, pset) in name["bond_name"].items():
            head, idx = name["bond"][bond_name]
            cluster_no = int(name["alias"][head].split("_")[1])
            prop_no = cluster_no - 1 if cluster_no < cluster_bond_n else cluster_bond_n - 1
            (c1, c2), width, opacity = default_style["bond"][prop_no]
            if n_pset > 1:
                cluster_name = f"B{cluster_no}({pset})"
            else:
                cluster_name = f"B{cluster_no}"
            _, mp, ij, _, _ = data["bond"][bond_name]

            bond_alias_name = name["bond"][bond_name][0]
            bond_no = int(bond_name.split("_")[1])
            _, _, nd, n = info["rep_bond"][bond_alias_name]
            color1, color2 = (c1, c1) if nd == "ND" else (c1, c2)
            if mode == "debug":
                label = f"b{bond_no} ({ij[0]+1},{ij[1]+1})"
            else:
                label = f"b{bond_no}/{n}th: " + self._mapping_str(mp)
            v, c = bond.split("@")
            opa = opacity * 0.5 if mode != "standard" else opacity
            qtdraw.plot_bond(c, v, color=color1, color2=color2, width=width * scale, opacity=opa, name=cluster_name, label=label)
            if mode != "standard":
                v = NSArray(v).transform(self.A)
                v_len = v.norm() * 0.25
                acolor = "red" if idx == 1 and pset == 1 else "black"
                qtdraw.plot_vector(
                    c,
                    v,
                    length=v_len,
                    width=width * scale * 0.3,
                    color=acolor,
                    opacity=opacity,
                    name=cluster_name,
                    label=label,
                )

        dic = qtdraw.create_dict(self.model)
        qtdraw.close()
        del qtdraw

        return dic

    # ==================================================
    def _set_rep_site(self, info):
        """
        set representative site.
        """
        spinful = info["spinful"]
        crystal = info["crystal"]

        rep_site = {}
        for name, (pos, orb_list) in info["site"].items():
            wp = str(self._mpm.group.find_wyckoff_position(str(pos)))
            pos = str(NSArray(pos))
            orb_list = split_orb_list_rank_block(
                sort_orb_list(parse_orb_list(orb_list, spinful, crystal), spinful, crystal), spinful, crystal
            )
            if not info["molecule"]:
                pos = str(self._mpm.group.shift_home_unit_cell(pos))
            rep_site[name] = (pos, wp, orb_list)

        info["rep_site"].update(rep_site)

    # ==================================================
    def _set_site(self, info, name, data):
        """
        set sites in cluster or cell.
        """
        alias = {}
        cluster_site = {}
        cell_site = {}
        data_site = {}
        name_site = {}
        site_to_name = {}
        orbital = {}

        if not info["molecule"]:
            plus_set = NSArray.from_str(data["plus_set"])
            n_pset = len(plus_set)
        else:
            n_pset = 1

        site_no = 0
        for no, (site, (pos, _, orb_list)) in enumerate(info["rep_site"].items()):
            if info["molecule"]:
                s_map = self._mpm.group.site_mapping(pos)
                basic_num = len(s_map)
            else:
                s_map = self._mpm.group.site_mapping(pos, plus_set=True)
                basic_num = len(s_map) // n_pset
            position = list(s_map.keys())
            prop = list(s_map.values())
            cluster = f"S_{no+1:03d}"
            alias[cluster] = site
            alias[site] = cluster
            cluster_site[cluster] = []
            for rsite_no in range(basic_num):
                s, mp = position[rsite_no], prop[rsite_no]  # as sorted by pset group in s_map.
                s_no = site_no + rsite_no
                sname = f"site_{s_no+1:03d}"
                cluster_site[cluster].append(sname)
                ij = (s_no, s_no)
                data_site[sname] = (s, mp, ij)
                if n_pset > 1:
                    for pset in range(n_pset):
                        ss = position[pset * basic_num + rsite_no]
                        cell_site[f"{site}_{rsite_no+1}({pset+1})"] = (ss, self._mapping_str(mp))
                else:
                    cell_site[f"{site}_{rsite_no+1}"] = (s, self._mapping_str(mp))
                name_site[sname] = (site, rsite_no + 1)
                orbital[sname] = orb_list
            assert data_site[f"site_{site_no+1:03d}"][1][0] == 0, f"first SO is not identity operation, {mp}"
            for rsite_no in range(basic_num):
                s_no = site_no + rsite_no
                sname = f"site_{s_no+1:03d}"
                for pset in range(n_pset):
                    s = position[pset * basic_num + rsite_no]
                    site_to_name[s] = (sname, pset + 1)
            site_no = site_no + basic_num

        # check orbitals
        d = {}
        spinful = info["spinful"]
        crystal = info["crystal"]
        for sname, orb_list in orbital.items():
            for orbs in orb_list:
                o = orbs[0]
                l = rank(o, spinful, crystal)
                if l in d:
                    d[l] += [(sname, orbs)]
                else:
                    d[l] = [(sname, orbs)]
        for l, lst in d.items():
            sname, orbs = lst[0]
            if len(lst) > 1:
                for sname_, orbs_ in lst[1:]:
                    if orbs != orbs_:
                        s1 = name_site[sname][0]
                        s2 = name_site[sname_][0]
                        raise Exception(f"invalid orbitals are given, {s1} {orbs} != {s2} {orbs_}.")

        info["cell_site"].update(cell_site)

        name["alias"].update(alias),
        name["site"].update(name_site)
        name["site_name"].update(site_to_name)

        data["cluster_site"].update(cluster_site)
        data["site"].update(data_site)

        return orbital

    # ==================================================
    def _set_orbital(self, info, name, orbital):
        ket_site = [f"{head}_{idx}" for head, idx in name["site"].values()]

        unique_orb = set()
        ket = []
        ket_dict = {}
        ket_no = 0
        for no, site in enumerate(name["site"].keys()):
            s = ket_site[no]
            for orb in sum(orbital[site], []):
                ket.append(f"{orb}@{s}")
                ket_dict[(orb, no)] = ket_no
                ket_no = ket_no + 1
                unique_orb.add(orb)

        dimension = len(ket)
        orbital = sort_orb_list(list(unique_orb), info["spinful"], info["crystal"])

        info["dimension"] = dimension
        info["ket"] = ket
        info["ket_site"] = ket_site
        info["orbital"] = orbital

        return ket_dict

    # ==================================================
    def _set_rep_bond(self, info, name, detail):
        """
        set representative bond.
        """
        rep_bond_all = {}
        for tail, head, neighbor in info["bond"]:
            if type(neighbor) is not list:
                neighbor = [neighbor]
            if tail not in info["rep_site"].keys():
                raise Exception(f"{tail} is not found in sites.")
            if head not in info["rep_site"].keys():
                raise Exception(f"{head} is not found in sites.")

            # create rep. bond.
            tail_pos = info["rep_site"][tail][0]
            head_pos = info["rep_site"][head][0]
            rep_bond = self._representative_bond(tail_pos, head_pos, info, detail)
            rep_bond_list = [{}]
            for n, lst in enumerate(rep_bond):
                n_rep_bond = {}
                for i, b in enumerate(lst):
                    bb = NSArray(b)
                    v, c = bb.convert_bond("bond")
                    wp = str(self._mpm.group.find_wyckoff_position(str(c)))
                    _, nd = self._mpm.group.bond_mapping(b)
                    nd = "ND" if nd else "D"
                    b = f"{v}@{c}"
                    b = self._rep_bond_site_order(b, info, name)
                    bond_name = f"{tail}:{head}:{n+1}:{i+1}"
                    n_rep_bond[bond_name] = (b, wp, nd, n + 1)
                rep_bond_list.append(n_rep_bond)

            bond_th = f"{tail}_{head}"
            rep_bond_all[bond_th] = rep_bond_list

            # set rep. bond for required neighbors.
            rep_bond_selected = {}
            for n in neighbor:
                b = rep_bond_all[bond_th][n]
                for k, v in b.items():
                    rep_bond_selected[k] = v

            info["rep_bond"].update(rep_bond_selected)

        detail["rep_bond_all"].update(rep_bond_all)

    # ==================================================
    def _rep_bond_site_order(self, bond, info, name):
        """
        get rep_bond in order of site numbers.
        """
        site_to_name = name["site_name"]

        if info["molecule"]:
            b_map, _ = self._mpm.group.bond_mapping(bond)
        else:
            b_map, _ = self._mpm.group.bond_mapping(bond, plus_set=True)

        lst = []
        for b in b_map.keys():
            bb = NSArray(b)
            t, h = bb.convert_bond("bond_th")
            v, c = bb.convert_bond("bond")
            if info["molecule"]:
                tail_name = site_to_name[str(t)][0]
                head_name = site_to_name[str(h)][0]
            else:
                tail_name = site_to_name[str(t.shift())][0]
                head_name = site_to_name[str(h.shift())][0]
            t_no, h_no = (int(tail_name.split("_")[1]) - 1, int(head_name.split("_")[1]) - 1)
            lst.append((t_no, h_no, b))
        lst = sorted(lst)

        return lst[0][2]

    # ==================================================
    def _set_bond(self, info, name, data):
        """
        set bonds in cell.
        """
        # set bond in cluster or cell for required neighbors.
        alias = {}
        cluster_bond = {}
        cell_bond = {}
        data_bond = {}
        name_bond = {}
        bond_to_name = {}

        site_to_name = name["site_name"]

        if not info["molecule"]:
            plus_set = NSArray.from_str(data["plus_set"])
            n_pset = len(plus_set)
        else:
            n_pset = 1

        bond_no = 1  # no=0 is for null vector.
        for no, (rbname, (bond, _, _, _)) in enumerate(info["rep_bond"].items()):
            if info["molecule"]:
                b_map, _ = self._mpm.group.bond_mapping(bond)
                basic_num = len(b_map)
            else:
                b_map, _ = self._mpm.group.bond_mapping(bond, plus_set=True)
                basic_num = len(b_map) // n_pset
            bonds = list(b_map.keys())
            prop = list(b_map.values())
            cluster = f"B_{no+1:03d}"
            alias[cluster] = rbname
            alias[rbname] = cluster
            cluster_bond[cluster] = []
            b_vector = {}
            for rbond_no in range(basic_num):
                b, mp = bonds[rbond_no], prop[rbond_no]  # as sorted by pset group in s_map.
                b_no = bond_no + rbond_no
                bname = f"bond_{b_no:03d}"
                cluster_bond[cluster].append(bname)
                bb = NSArray(b)
                t, h = bb.convert_bond("bond_th")
                v, c = bb.convert_bond("bond")
                if info["molecule"]:
                    tail_name = site_to_name[str(t)][0]
                    head_name = site_to_name[str(h)][0]
                else:
                    tail_name = site_to_name[str(t.shift())][0]
                    head_name = site_to_name[str(h.shift())][0]
                ht_no = (int(head_name.split("_")[1]) - 1, int(tail_name.split("_")[1]) - 1)
                assert ht_no[0] >= ht_no[1], f"site indices are wrong, {ht_no}"
                vc = f"{v}@{c}"
                data_bond[bname] = (vc, mp, ht_no, str(v), b)
                br = str(NSArray(vc).regular_direction()).split("@")[0]
                b_vector[br] = b_vector.get(br, []) + [bname]
                if n_pset > 1:
                    for pset in range(n_pset):
                        v, c = NSArray(bonds[pset * basic_num + rbond_no]).convert_bond("bond")
                        vc1 = f"{v}@{c}"
                        cell_bond[f"{rbname}_{rbond_no+1}({pset+1})"] = (vc1, self._mapping_str(mp))
                else:
                    cell_bond[f"{rbname}_{rbond_no+1}"] = (vc, self._mapping_str(mp))
                name_bond[bname] = (rbname, rbond_no + 1)
            assert data_bond[f"bond_{bond_no:03d}"][1][0] == 0, f"first SO is not identity operation, {mp}"
            for rbond_no in range(basic_num):
                b_no = bond_no + rbond_no
                bname = f"bond_{b_no:03d}"
                for pset in range(n_pset):
                    v, c = NSArray(bonds[pset * basic_num + rbond_no]).convert_bond("bond")
                    b = f"{v}@{c}"
                    bond_to_name[b] = (bname, pset + 1)
            bond_no = bond_no + basic_num

            # replace same bond vector.
            for lst in b_vector.values():
                top = lst[0]
                v0 = data_bond[top][3]
                for i in lst[1:]:
                    b, mp, ht_no, v, th = data_bond[i]
                    nv = top if v0 == v else "-" + top
                    data_bond[i] = (b, mp, ht_no, nv, th)
            data["bond"].update(data_bond)

        info["cell_bond"].update(cell_bond)

        name["alias"].update(alias)
        name["bond"].update(name_bond)
        name["bond_name"].update(bond_to_name)

        data["cluster_bond"].update(cluster_bond)
        data["bond"].update(data_bond)

    # ==================================================
    def _set_matrix(self, data, orbital, ket_dict):
        cluster_atomic = {}
        atomic_braket = {}

        ij_list = [v[2] for v in data["site"].values()] + [v[2] for v in data["bond"].values()]

        for ij in ij_list:
            i, j = ij
            site_i = f"site_{i+1:03d}"
            site_j = f"site_{j+1:03d}"
            cluster_atomic[ij] = []
            if i == j:
                orbs_i = orbital[site_i]
                orbs_j = orbs_i
                braket = sum([[(orb, orbs_j[j]) for j in range(i, len(orbs_i))] for i, orb in enumerate(orbs_i)], [])
            else:
                orbs_i = orbital[site_i]
                orbs_j = orbital[site_j]
                braket = sum([[(orbi, orbj) for orbj in orbs_j] for orbi in orbs_i], [])
            cluster_atomic[ij] = [(f"{bra}:{ket}", (bra[0], i), (ket[0], j)) for bra, ket in braket]
            for bra, ket in braket:
                atomic_braket[f"{bra}:{ket}"] = (bra, ket)

        rs = {k: f"M_{i+1:03d}" for i, k in enumerate(atomic_braket.keys())}
        d = {rs[k]: v for k, v in atomic_braket.items()}
        atomic_braket = d

        dca = {}
        for k, lst in cluster_atomic.items():
            nlst = [(ket_dict[k1], ket_dict[k2], rs[i]) for i, k1, k2 in lst]
            dca[k] = nlst
        cluster_atomic = dca

        data["cluster_atomic"].update(cluster_atomic)
        data["atomic_braket"].update(atomic_braket)

    # ==================================================
    @classmethod
    def _mapping_str(cls, mp):
        """
        modify SO mapping (+1).

        Args:
            mp (list): mapping list.

        Returns:
            str: modified mapping list.
        """
        m = [i - 1 if i < 0 else i + 1 for i in mp]
        return str(m).replace(" ", "")

    # ==================================================
    def _group_str(self, molecule):
        """
        create group string.

        Args:
            molecule (bool): molecule or crystal ?

        Returns:
            str: info. for group.
        """
        no, _, IS, setting = self._mpm.group.tag.info()
        if setting:
            setting = "(" + setting + ")"
        SS = str(self._mpm.group.tag)
        group_str = "point" if molecule else "space"
        ret = f"{group_str} group No. {no} : {SS} / {IS}"
        if setting:
            ret += f" {setting}"
        if not molecule:
            ret += f" : PG {self._mpm.point_group}"

        return ret

    # ==================================================
    def _site_grid(self, site, info, detail, repeat=False):
        """
        site grid.

        Args:
            site (str): representative site.
            info (dict): info. dict.
            detail (dict): detail dict.

        Returns:
            NSArray: all sites for given grid.
        """
        if info["molecule"]:
            site = self._mpm.group.transform_site(site, remove_duplicate=True)
            return site

        site = self._mpm.group.transform_site(site, shift=True, remove_duplicate=True, plus_set=True)

        if repeat:
            offset = detail["cell_range"][::2]
            N = [i - j for i, j in zip(detail["cell_range"][1::2], offset)]
            igrid = NSArray.igrid(N, offset)
            igrid = np.tile(igrid.numpy(), (len(site), 1))

            ns = N[0] * N[1] * N[2]
            all_site = NSArray(np.tile(site.numpy(), (ns, 1)) + igrid, "vector")
        else:
            all_site = site

        return all_site

    # ==================================================
    def _representative_bond(self, tail, head, info, detail):
        """
        generate representative bonds.

        Args:
            tail (str): tail position.
            head (str): head position.
            info (dict): info dict.
        Returns:
            list: list of representative bonds in each neighbor.
        """
        tail = NSArray("{" + tail + "}")
        head = self._site_grid(head, info, detail, repeat=True)
        all_bond = NSArray.distance(tail, head, self._G)
        if info["molecule"]:
            all_bond = list(all_bond.values())[1:]
        else:
            all_bond = list(all_bond.values())[1 : detail["max_neighbor"]]
        rep_bond = []
        for lst in all_bond:
            th = [f"{tail[0]};{head[i]}" for _, i in lst]
            bs = NSArray.from_str(th)
            rb = self._mpm.group.find_rep_bond(bs)
            rep_bond.append(rb.str())

        return rep_bond

    # ==================================================
    @property
    def A(self):
        return self._A

    # ==================================================
    @property
    def G(self):
        return self._G

    # ==================================================
    @property
    def A_norm(self):
        return self._A_norm

    # ==================================================
    @classmethod
    def regularize(cls, d):
        """
        regularize minimal model dict.

        Args:
            d (dict): dict of model.

        Returns:
            dict: regularized model dict.
        """
        # default setting.
        if "cell" not in d.keys():
            d["cell"] = None

        model = d["model"]

        if "option" not in d.keys():
            d["option"] = {"view": None, "view_mode": "standard", "output": model, "minimal_samb": True}
        else:
            if "view" not in d["option"].keys():
                d["option"]["view"] = None
            if "view_mode" not in d["option"].keys():
                d["option"]["view_mode"] = "standard"
            if "output" not in d["option"].keys():
                d["option"]["output"] = model
            if "minimal_samb" not in d["option"].keys():
                d["option"]["minimal_samb"] = True

        if "generate" not in d.keys():
            d["generate"] = {"model_type": "tight_binding", "time_reversal_type": "electric", "irrep": None}
        else:
            if "model_type" not in d["generate"].keys():
                d["generate"]["model_type"] = "tight_binding"
            if "time_reversal_type" not in d["generate"].keys():
                d["generate"]["time_reversal_type"] = "electric"
            if "irrep" not in d["generate"].keys():
                d["generate"]["irrep"] = None

        if "spinful" not in d.keys():
            d["spinful"] = False
        else:
            if d["generate"]["model_type"] == "phonon":
                d["spinful"] = False

        if "k_point" not in d.keys():
            d["k_point"] = {"Γ": "[0, 0, 0]", "X": "[1/2, 0, 0]"}
        if "k_path" not in d.keys():
            d["k_path"] = "Γ-X"

        return d

    # ==================================================
    @classmethod
    def _header(cls):
        return header_str



"""
WyckoffGSet manages a set of point/space-group wyckoff positions (conventional, reduced).
"""
from multipie.wyckoff.wyckoff_g import WyckoffG
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_list import TagList


# ==================================================
class WyckoffGSet(dict):  # dict of (group tag, wyckoff set), {TagGroup: WyckoffG}.
    """
    a set of point/space-group wyckoff positions.

    Attributes:
        tag (str): class name tag.
    """

    # ==================================================
    def __init__(self):
        """
        initialize the class.
        """
        self.tag = __class__.__name__
        """class name tag."""
        for tag in TagGroup.create():
            self[tag] = WyckoffG(str(tag))
        for tag in TagGroup.create(space_group=True):
            self[tag] = WyckoffG(str(tag))

    # ==================================================
    def __str__(self):
        return self.tag

    # ==================================================
    def __repr__(self):
        return self.tag

    # ==================================================
    def latex(self):
        return self.tag

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagGroup(tag))
        else:
            return self.get(tag)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())



"""
point-group information.
    point-group tag : (No, Schoenflies symbol, international symbol(short), setting, crystal)
"""
_data_tag_point_group = {
    "C1": (1, "C_{1}", "1", "", "triclinic"),
    "Ci": (2, "C_{i}", "-1", "", "triclinic"),
    "C2": (3, "C_{2}", "2", "b-axis", "monoclinic"),
    "Cs": (4, "C_{s}", "m", "b-axis", "monoclinic"),
    "C2h": (5, "C_{2h}", "2/m", "b-axis", "monoclinic"),
    "D2": (6, "D_{2}", "222", "", "orthorhombic"),
    "C2v": (7, "C_{2v}", "mm2", "", "orthorhombic"),
    "D2h": (8, "D_{2h}", "mmm", "", "orthorhombic"),
    "C4": (9, "C_{4}", "4", "", "tetragonal"),
    "S4": (10, "S_{4}", "-4", "", "tetragonal"),
    "C4h": (11, "C_{4h}", "4/m", "", "tetragonal"),
    "D4": (12, "D_{4}", "422", "", "tetragonal"),
    "C4v": (13, "C_{4v}", "4mm", "", "tetragonal"),
    "D2d": (14, "D_{2d}", "-42m", "-42m", "tetragonal"),
    "D2d-1": (14, "D_{2d}-1", "-4m2", "-4m2", "tetragonal"),
    "D4h": (15, "D_{4h}", "4/mmm", "", "tetragonal"),
    "C3": (16, "C_{3}", "3", "", "trigonal"),
    "C3i": (17, "C_{3i}", "-3", "", "trigonal"),
    "D3": (18, "D_{3}", "312", "312", "trigonal"),
    "D3-1": (18, "D_{3}-1", "321", "321", "trigonal"),
    "C3v": (19, "C_{3v}", "3m1", "3m1", "trigonal"),
    "C3v-1": (19, "C_{3v}-1", "31m", "31m", "trigonal"),
    "D3d": (20, "D_{3d}", "-31m", "-31m", "trigonal"),
    "D3d-1": (20, "D_{3d}-1", "-3m1", "-3m1", "trigonal"),
    "C6": (21, "C_{6}", "6", "", "hexagonal"),
    "C3h": (22, "C_{3h}", "-6", "", "hexagonal"),
    "C6h": (23, "C_{6h}", "6/m", "", "hexagonal"),
    "D6": (24, "D_{6}", "622", "", "hexagonal"),
    "C6v": (25, "C_{6v}", "6mm", "", "hexagonal"),
    "D3h": (26, "D_{3h}", "-6m2", "-6m2", "hexagonal"),
    "D3h-1": (26, "D_{3h}-1", "-62m", "-62m", "hexagonal"),
    "D6h": (27, "D_{6h}", "6/mmm", "", "hexagonal"),
    "T": (28, "T", "23", "", "cubic"),
    "Th": (29, "T_{h}", "m-3", "", "cubic"),
    "O": (30, "O", "432", "", "cubic"),
    "Td": (31, "T_{d}", "-43m", "", "cubic"),
    "Oh": (32, "O_{h}", "m-3m", "", "cubic"),
}


"""
point groups with complex character.
"""
_data_tag_point_group_complex = ("C4", "S4", "C4h", "C3", "C3i", "C6", "C3h", "C6h", "T", "Th")


"""
point groups with complex character.
"""
_data_tag_point_group_real = tuple([i for i in _data_tag_point_group if i not in _data_tag_point_group_complex])


"""
point groups categorized by crystal.
"""
_data_tag_point_group_crystal = {val[4]: [] for val in _data_tag_point_group.values()}
for tag, val in _data_tag_point_group.items():
    _data_tag_point_group_crystal[val[4]].append(tag)
_data_tag_point_group_crystal = {c: tuple(v) for c, v in _data_tag_point_group_crystal.items()}


"""
point groups (cubic series).
"""
_data_tag_point_group_cubic_series = (
    _data_tag_point_group_crystal["cubic"]
    + _data_tag_point_group_crystal["tetragonal"]
    + _data_tag_point_group_crystal["orthorhombic"]
    + _data_tag_point_group_crystal["monoclinic"]
    + _data_tag_point_group_crystal["triclinic"]
)


"""
point groups (hexagonal series).
"""
_data_tag_point_group_hexagonal_series = _data_tag_point_group_crystal["hexagonal"] + _data_tag_point_group_crystal["trigonal"]



"""
CharacterPGSet manages a set of point-group character tables.
"""
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_irrep import TagIrrep
from multipie.tag.tag_list import TagList
from multipie.character.character_pg import CharacterPG
from multipie.data.data_compatibility_relation import _data_compatibility_relation


# ==================================================
class CharacterPGSet(dict):  # dict of (point-group, characters), {TagGroup: CharacterPG}.
    """
    a set of point-group characters.

    Attributes:
        tag (str): class name tag.
    """

    # ==================================================
    def __init__(self):
        """
        initialize the class.
        """
        self.tag = __class__.__name__
        """class name tag."""
        for tag in TagGroup.create():
            self[tag] = CharacterPG(tag)

    # ==================================================
    def __str__(self):
        return self.tag

    # ==================================================
    def __repr__(self):
        return self.tag

    # ==================================================
    def latex(self):
        return self.tag

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagGroup(tag))
        else:
            return self.get(tag)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())

    # ==================================================
    @classmethod
    def compatibility_relation(cls):
        """
        compatibility relation.

        Returns:
            {TagGroup:[TagIrrep]}: compatibility relation table, (group tag, list of irreps.).
        """
        comp_rel = {}
        cr = _data_compatibility_relation

        for tag, ir in cr.items():
            ir = TagList.from_str(TagIrrep, ir)
            comp_rel[TagGroup(tag)] = ir

        return comp_rel



"""
This class manages Clebsch-Gordan coefficient in point group.
"""
import sympy as sp
from sympy.physics.quantum.cg import CG

from multipie.harmonics.harmonics_pg import HarmonicsPG
from multipie.tag.tag_group import TagGroup


# ==================================================
class ClebschGordanPG:
    """
    point-group Clebsch-Gordan coefficient.

    Attributes:
        tag (TagGroup): point-group tag.
        harmonics (HarmonicsPG): the class to manage harmonics.
    """

    # ==================================================
    def __init__(self, pg_tag):
        """
        initialize the class.

        Args:
            pg_tag (TagGroup or str, optional): point-group tag.
        """
        pg_tag = TagGroup(str(pg_tag))
        self.tag = pg_tag
        """point-group tag."""
        self.harmonics = HarmonicsPG(pg_tag)
        """the class to manage harmonics."""

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def _tag_info(self, tag):
        t = +1 if tag.t_type == "electric" else -1
        p = 0 if tag.i_type == "polar" else 1
        return t, p, tag.rank, tag.irrep, tag.mul, tag.comp

    # ==================================================
    def cg(self, tag1, tag2, tag):
        """
        Clebsch-Gordan (CG) coefficient for direct product of 2 irreps., < tag1,comp1;tag2,comp2|tag>.

        Args:
            tag1 (TagMultipole): tag1.
            tag2 (TagMultipole): tag2.
            tag (TagMultipole): tag.

        Returns:
            sympy; CG coefficient
        """
        t1, p1, l1, _, _, _ = self._tag_info(tag1)
        t2, p2, l2, _, _, _ = self._tag_info(tag2)
        t, p, l, _, _, _ = self._tag_info(tag)

        if t1 * t2 * t != 1:
            return sp.S(0)
        if (l1 + l2 - l + p1 + p2 - p) % 2 != 0:
            return sp.S(0)
        if l < abs(l1 - l2) or l > l1 + l2:
            return sp.S(0)

        for h in self.harmonics.select(rank=l1):
            if self._tag_info(h.tag)[1:] == self._tag_info(tag1)[1:]:
                u1 = h.u_matrix()

        for h in self.harmonics.select(rank=l2):
            if self._tag_info(h.tag)[1:] == self._tag_info(tag2)[1:]:
                u2 = h.u_matrix()

        for h in self.harmonics.select(rank=l):
            if self._tag_info(h.tag)[1:] == self._tag_info(tag)[1:]:
                u = h.u_matrix()

        s = sp.S(0)
        for i, m1 in enumerate(reversed(range(-l1, l1 + 1))):
            for j, m2 in enumerate(reversed(range(-l2, l2 + 1))):
                for k, m in enumerate(reversed(range(-l, l + 1))):
                    if m != m1 + m2:
                        continue
                    c = CG(l1, m1, l2, m2, l, m).doit()
                    if c != 0:
                        s += (-sp.I) ** (l1 + l2 - l) * c * sp.conjugate(u1[i] * u2[j]) * u[k]

        return sp.expand(s)



from dataclasses import dataclass, InitVar
import sympy as sp
from copy import copy


# ==================================================
@dataclass(frozen=True, order=True)
class Tag:
    """
    default Tag class.
    """

    tag_str: InitVar[str]

    # ==================================================
    def __post_init__(self, tag_str):
        if tag_str == "":
            return
        d = self._from_str(tag_str)
        for k, v in d.items():
            object.__setattr__(self, k, v)

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        """
        create tag from str. impliment explicitly in each subclass.

        Args:
            tag (str): str to be parsed.

        Returns:
            dict: field-value pairs.
        """
        raise NotImplementedError()

    # ==================================================
    def __str__(self):
        """
        default function to print. impliment explicitly in each subclass.
        """
        raise NotImplementedError()

    # ==================================================
    def latex(self):
        """
        default function to convert latex format. impliment explicitly in each subclass.
        """
        raise NotImplementedError()

    # ==================================================
    def symbol(self, commutative=False):
        return sp.Symbol(self.latex(), commutative=commutative)

    # ==================================================
    def replace(self, **kwargs):
        """
        replace field-value.

        Args:
            kwargs (dict): field-value pairs to replace.

        Retruns:
            Tag: replaced tag.
        """
        ret = copy(self)
        for k, v in kwargs.items():
            object.__setattr__(ret, k, v)

        return ret



"""
TagMultipole manages tag of multipole.
"""
from dataclasses import dataclass
from multipie.tag.tag import Tag
from multipie.tag.tag_list import TagList
from multipie.tag.tag_irrep import TagIrrep
from multipie.data.data_harmonics import (
    _data_harmonics_polar,
    _data_harmonics_axial,
)
from multipie.const import __def_dict__, is_magnetic, is_axial


# ==================================================
@dataclass(frozen=True, order=True)
class TagMultipole(Tag):
    """
    tag of multipole.
    """

    head: str = "Q"  # type of multipole, Q/G/T/M.
    m_type: str = (
        "h"  # type of multipole, h(harmonics)/a(atomic)/c(cluster)/s(site-cluster)/b(bond-cluster)/u(uniform)/k(structure).
    )
    rank: int = 0  # rank or l.
    irrep: str = ""  # irrep. of point group.
    mul: int = 0  # multiplicity (>0 if multiplicity is required).
    comp: int = -1  # component or m (-l <= m <= l).
    s: int = -1  # spinless(0)/spinful(1).
    k: int = -1  # 0/(-1,0,1).

    dim: int = 1

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        head = tag[0]
        if tag[1] in ("h", "a", "c", "s", "b", "u", "k"):
            m_type = tag[1]
            tag = tag[3:]
        else:
            m_type = ""
            tag = tag[2:]
        tag = tag.replace(")", "")
        if tag.count("|") > 0:
            a, b = tuple(tag.split("|"))
            s, k = tuple(b.split(","))
            s = int(s)
            k = int(k)
        elif m_type == "a":
            a = tag
            s, k = 0, 0
        else:  # harmonics
            a = tag
            s, k = -1, -1

        rank, irrep, mul, comp = tuple(a.split(","))
        rank = int(rank)
        mul = int(mul) if mul else 0
        comp = int(comp) if comp else -1

        if irrep == "":
            dim = 2 * rank + 1
        elif irrep.count("a") > 0 or irrep.count("b") > 0:
            dim = 1
        else:
            dim = __def_dict__["irrep_dim"][irrep[0]]

        d = {
            "head": head,
            "m_type": m_type,
            "rank": rank,
            "irrep": irrep,
            "mul": mul,
            "comp": comp,
            "s": s,
            "k": k,
            "dim": dim,
        }

        return d

    # ==================================================
    def __str__(self):
        head = self.head
        m_type = self.m_type
        rank = self.rank
        irrep = self.irrep
        mul = self.mul
        comp = self.comp
        s = self.s
        k = self.k

        dim = self.dim

        b1, b2 = "(", ")"

        ss = head + m_type + b1
        ss += str(rank) + "," + irrep + ","
        if mul > 0:
            ss += str(mul)
        ss += ","
        if irrep == "" or (irrep != "" and dim > 1):
            ss += str(comp)
        if s == 1:
            ss += "|" + str(s) + "," + str(k)
        ss += b2

        return ss

    # ==================================================
    def latex(self):
        h = r"\mathbb{" + self.head + "}"
        if self.m_type == "":
            h = r"\hat{" + h + "}"

        sup = "(" + self.m_type
        if self.irrep:
            sup += "," if self.m_type != "" else ""
            sup += TagIrrep(self.irrep).latex()
        if self.mul > 0:
            sup += "," + str(self.mul)
        sup += ")"

        sub = str(self.rank)
        if self.dim > 1:
            sub += "," + str(self.m)

        end = ""
        if self.s == 1:
            end += "(" + str(self.s) + "," + str(self.k) + ")"

        s = h
        if sub:
            s += "_{" + sub + "}"
        if sup:
            s += "^{" + sup + "}"
        if end:
            s += end

        return s

    # ==================================================
    def tag_irrep(self):
        """
        tag of irrep.

        Returns:
            TagIrrep: tag of irrep. or None (for spherical case).
        """
        return TagIrrep(self.irrep) if self.irrep else None

    # ==================================================
    @property
    def i_type(self):
        """
        parity type.

        Returns:
            str: parity type, (polar/axial).
        """
        return __def_dict__["head_i"][self.head]

    # ==================================================
    @property
    def t_type(self):
        """
        time-reversal type.

        Returns:
            str: time-reversal type, (electric/magnetic).
        """
        return __def_dict__["head_t"][self.head]

    # ==================================================
    @property
    def l(self):
        """
        rank of multipole.

        Returns:
            int: rank.
        """
        return self.rank

    # ==================================================
    @property
    def m(self):
        """
        component of multipole (:math:`-l\le m \le l` for spherical version).

        Returns:
            int: component.
        """
        return self.comp

    # ==================================================
    def is_complex(self):
        return "a" in self.irrep or "b" in self.irrep or (self.irrep == "" and self.m != 0)

    # ==================================================
    def is_axial(self):
        return is_axial(self.i_type)

    # ==================================================
    def is_magnetic(self):
        return is_magnetic(self.t_type)

    # ==================================================
    def to_harmonics(self):
        """
        convert tag to corresponding harmonics.

        Returns:
            TagMultipole: converted tag.
        """
        head = __def_dict__["head_harm"][self.head]
        return self.replace(head=head, m_type="h", s=-1, k=-1)

    # ==================================================
    def to_polar(self):
        """
        convert tag to corresponding polar one.

        Returns:
            TagMultipole: converted tag.
        """
        head = __def_dict__["to_polar"][self.head]
        return self.replace(head=head)

    # ==================================================
    def reverse_t_type(self):
        """
        time-reversal tag.

        Returns:
            TagMultipole: time-reversal tag.
        """
        head = __def_dict__["head_tr"][self.head]
        return self.replace(head=head)

    # ==================================================
    def reverse_i_type(self):
        """
        space-reversal tag.

        Returns:
            TagMultipole: space-reversal tag.
        """
        head = __def_dict__["head_ir"][self.head]
        return self.replace(head=head)

    # ==================================================
    @classmethod
    def create(cls, axial=False):
        """
        harmonics tags.

        Args:
            axial (bool, optional): axial ?

        Returns:
            TagList: tags.
        """
        if axial:
            d = _data_harmonics_axial.values()
        else:
            d = _data_harmonics_polar.values()

        tags = sum([list(pg.keys()) for pg in d], [])

        tags = TagList.from_str(TagMultipole, tags)

        return tags

    # ==================================================
    @classmethod
    def create_spherical(cls, head="Q", rank=0, mul=0, comp=1, s=-1, k=-1):
        """
        create tag for spherical (atomic) multipole.

        Args:
            head (str, optional): _description_. Defaults to "Q".
            rank (int, optional): _description_. Defaults to 0.
            mul (int, optional): _description_. Defaults to 0.
            comp (int, optional): _description_. Defaults to 1.
            s (int, optional): _description_. Defaults to -1.
            k (int, optional): _description_. Defaults to -1.

        Returns:
            _type_: _description_
        """
        ret = TagMultipole("")
        ret = ret.replace(
            head=head,
            rank=rank,
            mul=mul,
            comp=comp,
            s=s,
            k=k,
            m_type="a",
            dim=2 * rank + 1,
        )
        return ret



"""
TagIrrep manages tag of irreducible representation.
"""
from dataclasses import dataclass
from multipie.tag.tag import Tag
from multipie.tag.tag_list import TagList
from multipie.data.data_tag_irrep import _data_tag_irrep
from multipie.const import __def_dict__


# ==================================================
@dataclass(frozen=True, order=True)
class TagIrrep(Tag):
    """
    tag of irreducible representation.
    """

    tag: str = ""  # tag of irrep.
    dim: int = 1  # dimension.
    parity: str = ""  # null/g/u.
    cmplx: str = ""  # null/a/b.

    head: str = ""
    sub: str = ""
    sup: str = ""

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        head = tag[0]
        sub = tag[1:]
        sub = sub.replace("'", "")
        np = tag.count("'")
        sup = "'" * np

        if tag.count("a"):
            c = "a"
            d = 1
        elif tag.count("b"):
            c = "b"
            d = 1
        else:
            c = ""
            d = __def_dict__["irrep_dim"][tag[0]]
        p = ""
        if tag.count("g"):
            p = "g"
        elif tag.count("u"):
            p = "u"

        d = {"tag": tag, "dim": d, "parity": p, "cmplx": c, "head": head, "sub": sub, "sup": sup}

        return d

    # ==================================================
    def __str__(self):
        s = self.head + self.sub + self.sup
        return s

    # ==================================================
    def latex(self):
        sup = self.sup.replace("'", r"\prime")
        sub = self.sub
        if sub.count("a"):
            sup += "(a)"
            sub = sub.replace("a", "")
        elif sub.count("b"):
            sup += "(b)"
            sub = sub.replace("b", "")

        s = self.head
        if sub:
            s += "_{" + sub + "}"
        if sup:
            s += "^{" + sup + "}"
        return s

    # ==================================================
    @classmethod
    def create(cls):
        """
        irrep. tags for point groups.

        Returns:
            TagList: all irrep. tags.
        """
        return TagList.from_str(TagIrrep, _data_tag_irrep)



"""
This file provides Pauli matices.
"""
import sympy
from sympy.physics.wigner import wigner_3j


# ==================================================
"""
Pauli matrices
"""


def sm0():
    return sympy.Matrix([[1, 0], [0, 1]])


def smx():
    return sympy.Matrix([[0, 1], [1, 0]])


def smy():
    return sympy.Matrix([[0, -sympy.I], [sympy.I, 0]])


def smz():
    return sympy.Matrix([[1, 0], [0, -1]])


def ss0():
    return sympy.symbols(r"\sigma_{0}", commutative=False)


def ssx():
    return sympy.symbols(r"\sigma_{x}", commutative=False)


def ssy():
    return sympy.symbols(r"\sigma_{y}", commutative=False)


def ssz():
    return sympy.symbols(r"\sigma_{z}", commutative=False)


# ==================================================
def pauli_matrix(i=4, matrix=True):
    """
    Pauli matrices (σ0, σx, σy, σz)

    Args:
        i (int, optional): = 0(0), 1(x), 2(y), 3(z), other(all)
        matrix (bool, optional): Matrix ?

    Returns:
        Matrix or sympy or dict: matrix or symbol or dict { str: Matrix }
    """
    if matrix:
        sm = [sm0(), smx(), smy(), smz()]
        if i in (0, 1, 2, 3):
            return sm[i]
        else:
            return {"0": sm[0], "x": sm[1], "y": sm[2], "z": sm[3]}
    else:
        ss = [ss0(), ssx(), ssy(), ssz()]
        if i in (0, 1, 2, 3):
            return ss[i]
        else:
            return {"0": ss[0], "x": ss[1], "y": ss[2], "z": ss[3]}


# ==================================================
def pauli_matrix_n(s, n, s1, s2):
    """
    Pauli matrices (σ00, σ11, σ10, σ1-1)

    Args:
        s (int): 0/1
        n (int): 0/+1/-1
        s1 (sympy): ±1/2
        s2 (sympy): ±1/2

    Returns:
       sympy: matrix element
    """
    if s == 0 and n == 0:
        return 1 if s1 == s2 else 0
    elif s == 1:
        return sympy.Rational((-1) ** (s1 - 1 / 2)) * sympy.sqrt(6) * wigner_3j(1 / 2, 1 / 2, 1, -s1, s2, n)
    else:
        return 0



"""
Harmonics manages point-group harmonics.
"""
import sympy as sp
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_multipole import TagMultipole
from multipie.harmonics.util.equivalent_operator import equivalent_operator_from_poly


# ==================================================
class Harmonics:
    """
    point-group harmonics.

    Attributes:
        tag (TagMultipole): multipole tag.
    """

    #
    #    __v (dict): variable to replace (x,y,z).
    #    __def (NSArray): definition string (normalized).
    #    __ex (NSArray): expression string (abc polynomial).
    #    __umat (NSArray): unitary matrix from Olm to harmonics (2l+1) descending order in m.
    #
    # ==================================================
    def __init__(self, m_tag, def_ex, ex, u):
        """
        initialize the class.

        Args:
            m_tag (TagMultipole or str): multipole tag.
            def_ex (str): definition of harmonics in terms of tesseral harmonics.
            ex (str): cartesian expression.
            u (str): unitary matrix from Olm in descending order in m, [l,l-1,...,-l].
        """
        m_tag = TagMultipole(str(m_tag))
        self.tag = m_tag
        """multipole tag."""
        self.__v = dict(zip(["x", "y", "z"], NSArray.vector3d(self.tag.head).var.values()))

        self.__def = NSArray(def_ex)
        self.__ex = NSArray(ex)
        self.__umat = NSArray(u)

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def is_axial(self):
        """
        axial harmonics ?

        Returns:
            bool: axial harmonics ?
        """
        return self.tag.is_axial

    # ==================================================
    def is_complex(self):
        """
        complex harmonics ?

        Returns:
            bool: complex harmonics ?
        """
        return self.tag.is_complex

    # ==================================================
    def definition(self):
        """
        definition of harmonics in terms of linear combination of Clm and Slm.

        Returns:
            NSArray: definition of harmonics.
        """
        return self.__def

    # ==================================================
    def expression(self, v=None):
        """
        explicit expression of harmonics.

        Args:
            v (NSArray, optional): vector variable to replace "[x,y,z]".

        Returns:
            NSArray: polynomial expression of harmonics.

        Notes:
            - when v is None, default variable is used.
        """
        if v is None:
            v = self.__v
        else:
            v = dict(zip(["x", "y", "z"], v))

        ex = self.__ex.subs(v)
        return ex

    # ==================================================
    def u_matrix(self):
        """
        unitary matrix from (l,m) to the harmonics, :math:`X_{harm} = U^{\dagger}X_{lm}U`.

        Returns:
            NSArray: unitary matrix from (l,m) to the harmonics with (2l+1) components, descending order in m.
        """
        return self.__umat

    # ==================================================
    def to_vector(self):
        """
        cartesian vector for rank-1 expression.

        Returns:
            NSArray: converted cartesian vector.
        """
        assert self.tag.rank == 1, "rank is not 1."

        ex = self.__ex.subs(
            {
                "x": sp.Matrix([1, 0, 0]).T,
                "y": sp.Matrix([0, 1, 0]).T,
                "z": sp.Matrix([0, 0, 1]).T,
            }
        )
        return NSArray(ex[0], "vector")

    # ==================================================
    def equivalent_operator(self, j):
        """
        equivalent operator in descending order in Jz.

        Args:
            j (str): magnitude of angular momentum, J (0, 1/2, 1, ...).

        Returns:
            NSArray: equivalent-operator matrix.
        """
        poly = str(self.expression(v="[x,y,z]"))
        return equivalent_operator_from_poly(poly, j)

    # ==================================================
    @classmethod
    def create_equivalent_operator(cls, poly, j):
        """
        equivalent operator in descending order in Jz.

        Args:
            poly (str): (x,y,z) polynomial.
            j (str): magnitude of angular momentum, J (0, 1/2, 1, ...).

        Returns:
            NSArray: equivalent-operator matrix.
        """
        return equivalent_operator_from_poly(poly, j)



"""
TagWyckoff manages tag of wyckoff position.
"""
from dataclasses import dataclass
from multipie.tag.tag import Tag
from multipie.tag.tag_list import TagList
from multipie.data.data_tag_point_group import _data_tag_point_group
from multipie.data.data_wyckoff_pg import _data_wyckoff_pg
from multipie.data.data_wyckoff_sg import _data_wyckoff_sg


# ==================================================
@dataclass(frozen=True, order=True)
class TagWyckoff(Tag):
    """
    tag of wyckoff position.
    """

    letter: str = "a"  # Wyckoff letter.
    n: int = 1  # number of sites.

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        letter = tag[-1]
        n = int(tag[:-1])
        d = {"letter": letter, "n": n}
        return d

    # ==================================================
    def __str__(self):
        s = str(self.n) + self.letter
        return s

    # ==================================================
    def latex(self):
        return str(self)

    # ==================================================
    @classmethod
    def create(cls, tag):
        """
        point-group tags.

        Args:
            tag (str): point-group/space-group tag.

        Returns:
            [TagWyckoff]: tags.
        """
        if tag in _data_tag_point_group.keys():
            tags = TagList.from_str(TagWyckoff, [i[0] for i in _data_wyckoff_pg[tag]])
        else:
            tags = TagList.from_str(TagWyckoff, [i[0] for i in _data_wyckoff_sg[tag]])

        return tags



"""
SymmetryAdaptedModel manages symmetry adapted multipole basis set for cluster or crystal system.
"""
from itertools import product
import sympy as sp
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_list import TagList
from multipie.multipole.util.multipole_util import matrix_to_dict
from multipie.multipole.util.structure_samb_util import cs_list
from multipie.model.util.symmetry_adapted_model_util import (
    create_atomic_samb_set,
    create_cluster_samb_set,
    create_uniform_samb_set,
    create_structure_samb_set,
    create_z_samb_set,
    create_zk_samb_set,
    redefine_index,
)
from multipie import __version__


# ==================================================
header_str = """
=== SAMB (* only for crystal) ===
- info
    - atomic : { "M_#" : ["amp_#"] }
    - site_cluster : { "S_#" : ["smp_#"] }
    - bond_cluster : { "B_#" : ["bmp_#"] }
    - uniform : { "S_#"/"B_#" : ["ump_#"] }
    - structure* : { "B_#" : ["kmp_#"] }
    - Z : { ("M_#", "S_#"/"B_#") : ["z_#"] }
    - version : MultiPie version
    - harmonics : { head : [TagMultipole] }

- data
    - atomic : { "amp_#" : ( TagMultipole, shape, [(i, j, matrix element)] ) }
    - site_cluster : { "smp_#" : ( TagMultipole, [vector component] ) }
    - bond_cluster : { "bmp_#" : ( TagMultipole, [vector component] ) }
    - uniform : { "ump_#" : ( TagMultipole, shape, [(i, j, matrix element)] ) }
    - structure* : { "kmp_#" : (TagMultipole, "structure factor") }
    - Z : {"z_#" : ( TagMultipole, [(coeff, "amp_#", "smp_#"/"bmp_#/ump_#")] ) }
    - Zk* : {"z_#" : ( TagMultipole, [(coeff, "amp_#", "ump_#", "kmp_#")] ) }
"""

# ==================================================
matrix_header_str = """
== SAMB in full matrix form (* only for crystal) ===
- model : model name.
- molecule : molecule or crystal ?
- group : (tag, detailed str)
- dimension : dimension of full matrix
- ket : ket basis list, orbital@site
- version : MultiPie version
- k_point* : representative k points
- k_path* : high-symmetry line in k space
- A* : transform matrix, [a1,a2,a3]
- bond* : { "bond_#": "vector" }
- matrix : { "z_#": "matrix" }
"""


# ==================================================
def _run(msg, mpm, func, *args, **kwargs):
    mpm.log(f"  * {msg} ... ", None, end="")
    mpm.set_stamp()
    res = func(*args, **kwargs)
    mpm.log("done")

    return res


# ==================================================
class SymmetryAdaptedModel(dict):
    """
    symmetry adapted multipole basis set for molecule or crystal.
    """

    # Attributes:
    #     _model (dict): molecule or crystal model information.
    #     _parallel (bool): use parallel code.
    #     _verbose (bool): verbose parallel info.
    #
    # ==================================================
    def __init__(self, model, mpm, head, irrep, toroidal_priority=False):
        """
        initialize the class.

        Args:
            model (dict): model information of cluster or crystal system.
            mpm (MultiPieManager): MultiPie manager.
            head (list or str): heads of SAMB to be created.
            irrep (list or str): irreps. of SAMB to be created.
            toroidal_priority (bool): create toroidal multipoles (G,T) in priority? else prioritize conventional multipoles (Q,M).
        """
        self._model = model
        self._parallel = mpm.parallel
        self._verbose = mpm.verbose

        molecule = model["info"]["molecule"]
        spinful = model["info"]["spinful"]

        cluster_site = model["data"]["cluster_site"]
        site = model["data"]["site"]
        cluster_bond = model["data"]["cluster_bond"]
        bond = model["data"]["bond"]

        cluster_atomic = model["data"]["cluster_atomic"]
        atomic_braket = model["data"]["atomic_braket"]

        g = mpm.group
        pg = mpm.point_group

        # atomic
        func = create_atomic_samb_set
        args = [pg, atomic_braket, spinful, self._parallel]
        atomic_info, atomic_data = _run("atomic", mpm, func, *args)

        # site/bond-cluster
        rep_site_dict = {cluster_tag: site[site_list[0]][0] for cluster_tag, site_list in cluster_site.items()}
        rep_bond_dict = {cluster_tag: bond[bond_list[0]][0] for cluster_tag, bond_list in cluster_bond.items()}

        func = create_cluster_samb_set
        args = [g, rep_site_dict, rep_bond_dict, self._parallel]
        site_cluster_info, site_cluster_data, bond_cluster_info, bond_cluster_data = _run("site/bond-cluster", mpm, func, *args)

        # uniform
        cluster_samb_set = {S_i: [site_cluster_data[smp_i] for smp_i in lst] for S_i, lst in site_cluster_info.items()}
        cluster_samb_set |= {B_i: [bond_cluster_data[bmp_i] for bmp_i in lst] for B_i, lst in bond_cluster_info.items()}
        braket_indexes_dict = {
            cluster_tag: [site[site_tag][2] for site_tag in site_list] for cluster_tag, site_list in cluster_site.items()
        }
        braket_indexes_dict |= {
            cluster_tag: [bond[bond_tag][2] for bond_tag in bond_list] for cluster_tag, bond_list in cluster_bond.items()
        }
        dim = len(site)

        func = create_uniform_samb_set
        args = [cluster_samb_set, braket_indexes_dict, dim, self._parallel]
        uniform_info, uniform_data = _run("uniform", mpm, func, *args)

        # Z
        x_tag_dict = {(M_i, atomic_data[amp_i][0]): amp_i for M_i, lst in atomic_info.items() for amp_i in lst}
        if molecule:
            y_tag_dict = {(SB_i, uniform_data[ump_i][0]): ump_i for SB_i, lst in uniform_info.items() for ump_i in lst}
        else:
            y_tag_dict = {(S_i, site_cluster_data[smp_i][0]): smp_i for S_i, lst in site_cluster_info.items() for smp_i in lst}
            y_tag_dict |= {(B_i, bond_cluster_data[bmp_i][0]): bmp_i for B_i, lst in bond_cluster_info.items() for bmp_i in lst}

        cluster_site_bond = cluster_site | cluster_bond
        braket_site_no_dict = {S_i: site[site_list[0]][2] for S_i, site_list in cluster_site.items()}
        braket_site_no_dict |= {B_i: bond[bond_list[0]][2] for B_i, bond_list in cluster_bond.items()}
        M_SB_list = []
        for SB_i in cluster_site_bond.keys():
            for _, _, M_i in cluster_atomic[braket_site_no_dict[SB_i]]:
                M_SB_list.append((M_i, SB_i))

        func = create_z_samb_set
        args = [g, x_tag_dict, y_tag_dict, M_SB_list, atomic_braket, toroidal_priority, self._parallel]
        z_info, z_data = _run("Z", mpm, func, *args, head=head, irrep=irrep)

        if not molecule:
            # structure
            bc_samb_set = {B_i: [bond_cluster_data[bmp_i] for bmp_i in lst] for B_i, lst in bond_cluster_info.items()}

            func = create_structure_samb_set
            args = [bc_samb_set, cluster_bond, bond, self._parallel]
            structure_info, structure_data = _run("structure", mpm, func, *args)

            # Zk
            bc_samb_set = {B_i: [(bmp_i, bond_cluster_data[bmp_i][1]) for bmp_i in lst] for B_i, lst in bond_cluster_info.items()}
            u_samb_set = [(ump_i, m) for ump_i, (_, m) in uniform_data.items()]
            k_samb_set = [(kmp_i, fk) for kmp_i, (_, fk) in structure_data.items()]

            func = create_zk_samb_set
            args = [z_data, bc_samb_set, u_samb_set, k_samb_set, cluster_bond, bond, dim, self._parallel]
            zk_data = _run("Zk", mpm, func, *args)

        # delete unused samb and redefine the serial number of multipoles.
        if model["info"]["option"]["minimal_samb"]:
            atomic_info_, atomic_data_, z_data = redefine_index(atomic_info, atomic_data, z_data, molecule=molecule)
            site_cluster_info, site_cluster_data, z_data = redefine_index(
                site_cluster_info, site_cluster_data, z_data, molecule=molecule
            )
            if len(site_cluster_info) > 0:
                init_idx = int(list(site_cluster_info.values())[-1][-1].split("_")[1]) + 1
            else:
                init_idx = 1
            bond_cluster_info, bond_cluster_data, z_data = redefine_index(
                bond_cluster_info, bond_cluster_data, z_data, molecule=molecule, init_idx=init_idx
            )
            uniform_info_, uniform_data_, z_data = redefine_index(uniform_info, uniform_data, z_data, molecule=molecule)
            if molecule:
                atomic_info, atomic_data = atomic_info_, atomic_data_
                uniform_info, uniform_data = uniform_info_, uniform_data_
            if not molecule:
                atomic_info, atomic_data, zk_data = redefine_index(atomic_info, atomic_data, zk_data)
                uniform_info, uniform_data, zk_data = redefine_index(uniform_info, uniform_data, zk_data)
                structure_info, structure_data, zk_data = redefine_index(structure_info, structure_data, zk_data)

        # harmonics
        head_list = ["Q", "G", "M", "T"]
        harmonics_info = {"Q": [], "G": []}
        for X in head_list:
            data = atomic_data | site_cluster_data | bond_cluster_data | uniform_data
            if not molecule:
                data |= structure_data
            tag_list = TagList([tag for _, (tag, _) in data.items()]).select(head=X)
            harmonics_list = sorted(set([tag.to_harmonics() for tag in tag_list]))

            X = "G" if X == "M" else X
            X = "Q" if X == "T" else X
            harmonics_info[X] += [str(htag) for htag in harmonics_list]

        harmonics_info = {"Q": sorted(set(harmonics_info["Q"])), "G": sorted(set(harmonics_info["G"]))}

        # info and data
        self["info"] = {
            "atomic": atomic_info,
            "site_cluster": site_cluster_info,
            "bond_cluster": bond_cluster_info,
            "uniform": uniform_info,
            "Z": z_info,
            "version": __version__,
        }
        self["data"] = {
            "atomic": atomic_data,
            "site_cluster": site_cluster_data,
            "bond_cluster": bond_cluster_data,
            "uniform": uniform_data,
            "Z": z_data,
        }

        if not molecule:
            self["info"]["structure"] = structure_info
            self["data"]["structure"] = structure_data
            self["data"]["Zk"] = zk_data

        self["info"]["harmonics"] = harmonics_info

    # ==================================================
    def create_dict(self):
        """
        create dictionary containing information and data of atomic, site-cluster, bond-cluster,
        uniform, (structure), Z(combined), Zk(fourier transform of Z).

        Returns:
            dict: information and data of multipole bases.
        """
        info = self["info"]

        atomic_data = {amp_i: (str(tag), *matrix_to_dict(m)) for amp_i, (tag, m) in self["data"]["atomic"].items()}
        site_cluster_data = {smp_i: (str(tag), str(v)) for smp_i, (tag, v) in self["data"]["site_cluster"].items()}
        bond_cluster_data = {bmp_i: (str(tag), str(v)) for bmp_i, (tag, v) in self["data"]["bond_cluster"].items()}
        uniform_data = {ump_i: (str(tag), *matrix_to_dict(m)) for ump_i, (tag, m) in self["data"]["uniform"].items()}
        z_data = {z_i: (str(tag), [tuple(map(str, v)) for v in lst]) for z_i, (tag, lst) in self["data"]["Z"].items()}

        data = {
            "atomic": atomic_data,
            "site_cluster": site_cluster_data,
            "bond_cluster": bond_cluster_data,
            "uniform": uniform_data,
            "Z": z_data,
        }
        if not self._model["info"]["molecule"]:
            structure_data = {
                kmp_i: (str(tag), str(fk).replace("'", "")) for kmp_i, (tag, fk) in self["data"]["structure"].items()
            }
            zk_data = {z_i: (str(tag), [tuple(map(str, v)) for v in lst]) for z_i, (tag, lst) in self["data"]["Zk"].items()}
            data["structure"] = structure_data
            data["Zk"] = zk_data

        return {"info": info, "data": data}

    # ==================================================
    def create_matrix(self, full=False, fmt="sympy"):
        """
        convert combined multipoles from dictionary to numerical matrix form (full matrix).

        Args:
            full (bool, optional): full matrix format ?
            fmt (str, optional): sympy/value.

        Returns:
            dict: matrix form of combined multipoles, z_dict/z_full.

        Note:
            z_dict = { "z_#" : [(i, j, ""/c###/s###, v)] }
            z_full = { "z_#" : NSArray(matrix) }
        """
        if fmt not in ["sympy", "value"]:
            raise KeyError(f"unknown format = {fmt} is given.")

        molecule = self._model["info"]["molecule"]
        cluster_atomic = self._model["data"]["cluster_atomic"]
        atomic_braket = self._model["data"]["atomic_braket"]
        dim_full = self._model["info"]["dimension"]

        atomic_data = self["data"]["atomic"]
        uniform_data = self["data"]["uniform"]
        if molecule:
            z_info, z_data = self["info"]["Z"], self["data"]["Z"]
        else:
            structure_data = self["data"]["structure"]
            z_info, z_data = self["info"]["Z"], self["data"]["Zk"]

        z_samb = {}
        z_samb["model"] = self._model["info"]["model"]
        z_samb["molecule"] = self._model["info"]["molecule"]
        z_samb["group"] = self._model["info"]["group"]
        z_samb["dimension"] = self._model["info"]["dimension"]
        z_samb["ket"] = self._model["info"]["ket"]
        z_samb["version"] = __version__

        if not molecule:
            z_samb["k_point"] = self._model["info"]["k_point"]
            z_samb["k_path"] = self._model["info"]["k_path"]
            z_samb["A"] = self._model["detail"]["A"]
            bond = {}
            for name, ex in self._model["data"]["bond"].items():
                ex = ex[0].split("@")[0]
                bond[name] = ex
            z_samb["bond"] = bond

        z_full = {}
        for (_, M_i, _), z_i_list in z_info.items():
            bra_orb_list, ket_orb_list = atomic_braket[M_i]
            dim_r, dim_c = len(bra_orb_list), len(ket_orb_list)
            for z_i in z_i_list:
                lst1 = z_data[z_i][1]
                Z = sp.Matrix.zeros(dim_full, dim_full)
                for v in lst1:
                    if len(v) == 3:
                        c, amp_i, ump_i = v
                        F = 1
                    else:
                        c, amp_i, ump_i, kmp_i = v
                        F = structure_data[kmp_i][1]
                    X = atomic_data[amp_i][1]
                    U = uniform_data[ump_i][1]
                    for (bra_site_no, ket_site_no), lst2 in cluster_atomic.items():
                        u = U[bra_site_no, ket_site_no]
                        if u != 0:
                            for bra_no, ket_no, M_i_ in lst2:
                                if M_i_ != M_i:
                                    continue
                                m = c * u * F * sp.Matrix(X)
                                Z[bra_no : bra_no + dim_r, ket_no : ket_no + dim_c] += m
                if not Z.is_hermitian:
                    Z = Z + Z.adjoint()
                z_full[z_i] = sp.expand(Z)

        DIGIT = 14
        if full:
            if fmt == "sympy":
                z_full = {z_i: str(mat.tolist()) for z_i, mat in z_full.items()}
            else:
                z_full_val = {}
                for z_i, Z in z_full.items():
                    Z_val = sp.Matrix.zeros(dim_full, dim_full)
                    for i, j in product(range(dim_full), range(dim_full)):
                        v1 = Z[i, j]
                        if v1 != 0:
                            if len(z_data[z_i][1][0]) == 3:
                                c = complex(v1)
                                c = round(c.real, DIGIT) + round(c.imag, DIGIT) * 1j
                                Z_val[i, j] += c
                            else:
                                d = {csn: v1.coeff(csn) for csn in cs_list(v1, real=True)}
                                for csn, c in d.items():
                                    if c != sp.S(0):
                                        c = complex(c)
                                        c = round(c.real, DIGIT) + round(c.imag, DIGIT) * 1j
                                        Z_val[i, j] += c * csn
                    z_full_val[z_i] = Z_val
                z_full = {z_i: str(NSArray(mat.tolist(), "matrix")) for z_i, mat in z_full_val.items()}
            z_samb["matrix"] = z_full
            return z_samb
        else:
            z_dict = {}
            for z_i, Z in z_full.items():
                mat = []
                for i, j in product(range(dim_full), range(dim_full)):
                    v1 = Z[i, j]
                    if v1 != 0:
                        if len(z_data[z_i][1][0]) == 3:
                            mat.append((i, j, "", str(v1)))
                        else:
                            d = {str(csn): v1.coeff(csn) for csn in cs_list(v1, real=True)}
                            for csn_str, c in d.items():
                                if c != sp.S(0):
                                    if fmt == "value":
                                        c = complex(c)
                                        c = round(c.real, DIGIT) + round(c.imag, DIGIT) * 1j
                                    mat.append((i, j, csn_str, str(c)))
                z_dict[z_i] = mat
            z_samb["matrix"] = z_dict
            return z_samb

    # ==================================================
    @classmethod
    def _header(cls):
        return header_str

    # ==================================================
    @classmethod
    def _matrix_header(cls):
        return matrix_header_str



"""
BaseAtomicMultipoleDataset manages atomic-multipole dataset for all four spinful basis.
"""
from multipie.multipole.base.base_atomic_multipole_set import BaseAtomicMultipoleSet


# ==================================================
class BaseAtomicMultipoleDataset(dict):  # dict of (basis-type, multipole set) { str: BaseAtomicMultipoleSet }.
    """
    atomic-multipole dataset for all four basis.

    Attributes:
        tag (str): class name tag.
    """

    # ==================================================
    def __init__(self):
        """
        initialize the class.
        """
        self.tag = __class__.__name__
        """class name tag."""

        b_type_lst = ["lm", "jm", "cubic", "hexagonal"]
        for b_type in b_type_lst:
            self[b_type] = BaseAtomicMultipoleSet(b_type)

    # ==================================================
    def __str__(self):
        return self.tag

    # ==================================================
    def __repr__(self):
        return self.tag

    # ==================================================
    def latex(self):
        return self.tag



"""
MaterialModel manages model information of cluster or crystal system.
"""
import datetime
import sympy as sp
from gcoreutils.nsarray import NSArray
from gcoreutils.pdf_via_latex import PDFviaLaTeX
from gcoreutils.convert_util import sympy_to_latex
from multipie.tag.tag_multipole import TagMultipole
from multipie.tag.tag_list import TagList
from multipie.model.material_model import MaterialModel
from multipie.model.util.atomic_matrix_util import convert_to_matrix
from multipie.multipole.util.atomic_orbital_util import to_latex
from multipie import __version__


# ==================================================
class ModelPDF:
    # ==================================================
    def vspace(self, pdf, sp="5mm"):
        pdf.text(r"\vspace{" + sp + "}" + "\n")

    # ==================================================
    def hr(self, pdf):
        pdf.text()
        pdf.text(r" \hfil \hrule height 1mm width \textwidth \hfil" + "\n")

    # ==================================================
    def ubar_name(self, name):
        if name.count("_") > 0:
            head, no = name.split("_")
            name = head + "$_{" + str(int(no)) + "}$"
        return name

    # ==================================================
    def comment(self, pdf, text):
        pdf.text("%" * 120)
        pdf.text("% " + text)
        pdf.text("%" * 120)

    # ==================================================
    def __init__(self, model_dict, samb_dict, mpm):
        """
        create pdf file from model and basis.

        Args:
            model_dict (_type_): _description_
            samb_dict (_type_): _description_
            mpm (_type_): _description_
        """
        self._mpm = mpm
        filename = mpm.filename(model_dict["info"]["model"] + "_samb", full=False)
        pdf = PDFviaLaTeX(filename, landscape=True, english=True, dir=mpm.dirname)

        info = model_dict["info"]
        name = model_dict["name"]
        data = model_dict["data"]

        cluster = samb_dict["info"]["site_cluster"] | samb_dict["info"]["bond_cluster"]
        samb_dict["info"]["cluster"] = cluster
        cluster = samb_dict["data"]["site_cluster"] | samb_dict["data"]["bond_cluster"]
        samb_dict["data"]["cluster"] = cluster

        pdf.text(r"\setlength{\baselineskip}{16pt}")
        pdf.text(r"\footnotesize")

        # main.
        self.header(pdf, info, name, data)
        self.vspace(pdf, "1cm")
        self.hr(pdf)
        pdf.text(r"\begin{itemize}")
        self.information(pdf, info, name, data)
        self.hr(pdf)
        if not info["molecule"]:
            self.lattice_info(pdf, info, name, data)
            self.hr(pdf)
        self.ket_space(pdf, info, name, data)
        self.hr(pdf)
        self.basis_info(pdf, info, data, samb_dict)
        self.hr(pdf)
        self.harmonics_info(pdf, info, samb_dict)
        self.hr(pdf)
        self.group_info(pdf, info, name, data)
        pdf.text(r"\end{itemize}")

        # postscript.
        mpm.log(f"  * wrote '{filename}.tex.'", None)
        mpm.log(f"  * wrote '{filename}.pdf.'", None)

        del pdf

    # ==================================================
    def header(self, pdf, info, name, data):
        dt = str(datetime.datetime.now())
        dt = dt[: dt.rfind(":")]

        # title.
        name = info["model"]
        pdf.title(r"SAMB for ``\texttt{" + f"{name}" + "}''")
        pdf.text(r"\begin{flushright}")
        pdf.text("Generated on " + dt + f" by MultiPie {__version__}")
        pdf.text(r"\end{flushright}")

    # ==================================================
    def information(self, pdf, info, name, data):
        g_tag = self._mpm.group.tag
        pdf.text(r"\item Group: " + g_tag.info(latex=True) + "\n")
        if not info["molecule"]:
            gp = self._mpm.point_group.tag
            pdf.text(r"\item Associated point group: " + gp.info(latex=True) + "\n")

        self.vspace(pdf)
        self.comment(pdf, "generation condition.")
        pdf.text(r"\item Generation condition" + "\n")
        pdf.text(r"\quad - model type: \texttt{" + info["generate"]["model_type"] + "}\n")
        pdf.text(r"\quad - time-reversal type: \texttt{" + info["generate"]["time_reversal_type"] + "}\n")
        pdf.text(r"\quad - irrep: \texttt{[" + ", ".join(info["generate"]["irrep"]) + "]}\n")
        spinful_str = "spinful" if info["spinful"] else "spinless"
        pdf.text(r"\quad - \texttt{" + spinful_str + "}\n")

    # ==================================================
    def ket_space(self, pdf, info, name, data):
        self.comment(pdf, "ket hilbert space.")
        pdf.text(r"\item Kets: dimension = " + str(len(info["ket"])))
        orb = to_latex([i.split("@")[0] for i in info["ket"]], info["spinful"], info["crystal"])
        s = [i.split("@")[1] for i in info["ket"]]
        ket = [i + "@" + j for i, j in zip(orb, s)]
        ket = [
            "$"
            + i.split("@")[0].replace("U", r"\uparrow").replace("D", r"\downarrow")
            + "$"
            + "@"
            + self.ubar_name(i.split("@")[1])
            for i in ket
        ]
        ket = sum([[str(no + 1), i] for no, i in enumerate(ket)], [])
        ncol = 5
        if len(ket) < 2 * ncol:
            ncol = len(ket) // 2
        pdf.table(
            ket,
            [""],
            ["No.", "ket"] * ncol,
            cpos="c" + "|cc" * ncol,
            caption="Hilbert space for full matrix.",
            stretch=1.3,
            long=True,
        )

        pdf.text()
        self.comment(pdf, "sites in unit cell.")
        pdf.text(r"\item Sites in (primitive) unit cell:")
        row = []
        tbl = []
        for cluster, lst in data["cluster_site"].items():
            cluster_name = self.ubar_name(cluster)
            row.append(cluster_name)
            tbl1 = []
            for site in lst:
                pos, mp, _ = data["site"][site]
                head, idx = name["site"][site]
                label = [
                    f"{head}$_{idx}$",
                    "$" + NSArray(pos).semi_evalf().latex() + "$",
                    MaterialModel._mapping_str(mp),
                ]
                tbl1 = tbl1 + label
            tbl.append(tbl1)
        pdf.table(
            tbl,
            row,
            ["site", "position", "mapping"],
            cpos="cc|c|l",
            caption="Site-clusters.",
            stretch=1.3,
            long=True,
            hl=True,
        )

        pdf.text()
        self.comment(pdf, "bonds in unit cell.")
        pdf.text(r"\item Bonds in (primitive) unit cell:")
        row = []
        tbl = []
        for cluster, lst in data["cluster_bond"].items():
            cluster_name = self.ubar_name(cluster)
            row.append(cluster_name)
            tbl1 = []
            for bond in lst:
                pos, mp, th, _, _ = data["bond"][bond]
                head_t, idx_t = name["site"][f"site_{th[0]+1:03d}"]
                t = head_t + "$_{" + str(idx_t) + "}$"
                head_h, idx_h = name["site"][f"site_{th[1]+1:03d}"]
                h = head_h + "$_{" + str(idx_h) + "}$"
                head, idx = name["bond"][bond]
                _, _, n, idx = head.split(":")
                no = int(bond.split("_")[1])
                label = [
                    r"b$_{" + str(no) + "}$",
                    t,
                    h,
                    str(int(n)),
                    str(int(idx.split("-")[0])),
                    "$" + NSArray(pos).semi_evalf().latex() + "$",
                    MaterialModel._mapping_str(mp),
                ]
                tbl1 = tbl1 + label
            tbl.append(tbl1)
        pdf.table(
            tbl,
            row,
            ["bond", "tail", "head", "$n$", "\#", r"$\bm{b}@\bm{c}$", "mapping"],
            caption="Bond-clusters.",
            cpos="cc|cc|c|c|c|l",
            stretch=1.3,
            long=True,
            hl=True,
        )

    # ==================================================
    def basis_info(self, pdf, info, data, samb):
        # SAMB.
        self.comment(pdf, "SAMB.")
        pdf.text(r"\item SAMB:" + "\n")

        Zinfo = {}
        for k, v in samb["info"]["Z"].items():
            for z_tag in v:
                Zinfo[z_tag] = k
        Zdict = samb["data"]["Z"]

        if not info["molecule"]:
            Zkdict = samb["data"]["Zk"]
            for no, (ztag, zj, zkj) in enumerate(zip(Zdict.keys(), Zdict.values(), Zkdict.values())):
                # Z basis.
                tag = TagMultipole(zj[0]).latex()
                _, m_block, c_block = Zinfo[ztag]
                pdf.text(r"\vspace{4mm}")
                pdf.text(
                    r"\noindent \fbox{No. {"
                    + str(no + 1)
                    + r"}} $\,\,\,"
                    + tag
                    + "$"
                    + f" [{self.ubar_name(m_block)},"
                    + r"\,"
                    + f"{self.ubar_name(c_block)}]"
                )
                eq = sp.S(0)
                rdict = {}
                for c, at, ct in zj[1]:
                    a_no = str(int(at.split("_")[1]))
                    c_no = str(int(ct.split("_")[1]))
                    rdict["a" + a_no] = TagMultipole(samb["data"]["atomic"][at][0]).latex()
                    rdict["c" + c_no] = TagMultipole(samb["data"]["cluster"][ct][0]).latex()
                    eq += (
                        NSArray(c)
                        * sp.symbols(r"A_{" + a_no + "}[a" + a_no + "]")
                        * sp.symbols(r"B_{" + c_no + "}[c" + c_no + "]")
                    )
                eq = sp.latex(eq).replace("A", r"\mathbb{X}").replace("B", r"\otimes\mathbb{Y}")
                for r, s in rdict.items():
                    eq = eq.replace(r, s)
                eq = r"\hat{\mathbb{Z}}_{" + str(no + 1) + "}=" + eq
                pdf.equation(eq, long=True)
                # Z(k) basis.
                eqk = sp.S(0)
                rdict = {}
                for term in zkj[1]:
                    if len(term) == 4:
                        c, at, ut, ft = term
                    else:
                        c, at, ut = term
                        ft = ""
                    a_no = str(int(at.split("_")[1]))
                    u_no = str(int(ut.split("_")[1]))
                    rdict["a" + a_no] = TagMultipole(samb["data"]["atomic"][at][0]).latex()
                    a_symbol = sp.symbols(r"A_{" + a_no + "}[a" + a_no + "]")
                    rdict["u" + u_no] = TagMultipole(samb["data"]["uniform"][ut][0]).latex()
                    u_symbol = sp.symbols(r"B_{" + u_no + "}[u" + u_no + "]")
                    if ft:
                        f_no = str(int(ft.split("_")[1]))
                        f_symbol = sp.symbols(r"C_{" + f_no + "}[f" + f_no + "]")
                        rdict["f" + f_no] = TagMultipole(samb["data"]["structure"][ft][0]).latex()
                    else:
                        f_symbol = sp.S(1)
                    eqk += NSArray(c) * a_symbol * u_symbol * f_symbol
                eqk = (
                    sp.latex(eqk)
                    .replace("A", r"\mathbb{X}")
                    .replace("B", r"\otimes\mathbb{U}")
                    .replace("C", r"\otimes\mathbb{F}")
                )
                for r, s in rdict.items():
                    eqk = eqk.replace(r, s)
                eqk = r"\hat{\mathbb{Z}}_{" + str(no + 1) + r"}(\bm{k})=" + eqk
                pdf.equation(eqk, long=True)
        else:
            for no, (ztag, zj) in enumerate(zip(Zdict.keys(), Zdict.values())):
                # Z basis.
                tag = TagMultipole(zj[0]).latex()
                _, m_block, c_block = Zinfo[ztag]
                pdf.text(r"\vspace{4mm}")
                pdf.text(
                    r"\noindent \fbox{No. {"
                    + str(no + 1)
                    + r"}} $\,\,\,"
                    + tag
                    + "$"
                    + f" [{self.ubar_name(m_block)},"
                    + r"\,"
                    + f"{self.ubar_name(c_block)}]"
                )
                eq = sp.S(0)
                rdict = {}
                for c, at, ct in zj[1]:
                    a_no = str(int(at.split("_")[1]))
                    c_no = str(int(ct.split("_")[1]))
                    rdict["a" + a_no] = TagMultipole(samb["data"]["atomic"][at][0]).latex()
                    rdict["u" + c_no] = TagMultipole(samb["data"]["uniform"][ct][0]).latex()
                    eq += (
                        NSArray(c)
                        * sp.symbols(r"A_{" + a_no + "}[a" + a_no + "]")
                        * sp.symbols(r"B_{" + c_no + "}[u" + c_no + "]")
                    )
                eq = sp.latex(eq).replace("A", r"\mathbb{X}").replace("B", r"\otimes\mathbb{U}")
                for r, s in rdict.items():
                    eq = eq.replace(r, s)
                eq = r"\hat{\mathbb{Z}}_{" + str(no + 1) + "}=" + eq
                pdf.equation(eq, long=True)

        # atomic SAMB group.
        self.comment(pdf, "atomic SAMB group.")
        row = []
        tbl = []
        for m_group, (bra, ket) in data["atomic_braket"].items():
            m_group = self.ubar_name(m_group)
            bra = to_latex(bra, info["spinful"], info["crystal"])
            ket = to_latex(ket, info["spinful"], info["crystal"])
            bra = "$" + ", ".join(bra).replace("U", r"\uparrow").replace("D", r"\downarrow") + "$"
            ket = "$" + ", ".join(ket).replace("U", r"\uparrow").replace("D", r"\downarrow") + "$"
            row.append(m_group)
            tbl.append([bra, ket])
        pdf.table(
            tbl,
            row,
            ["bra", "ket"],
            rc="group",
            caption="Atomic SAMB group.",
            cpos="c|c|c",
            stretch=1.3,
            long=True,
        )

        # atomic SAMB.
        self.comment(pdf, "atomic SAMB.")
        row = []
        tbl = []
        hl_no = -1
        hl = []
        for m_group, m_lst in samb["info"]["atomic"].items():
            hl.append(hl_no)
            m_group = self.ubar_name(m_group)
            for amp in m_lst:
                tag, shape, mat = samb["data"]["atomic"][amp]
                tag = "$" + TagMultipole(tag).latex() + "$"
                no = str(int(amp.split("_")[1]))
                no = "\mathbb{X}_{" + no + "}"
                mat = "$" + convert_to_matrix(shape[0], shape[1], mat).latex() + "$"
                row.append(no)
                tbl.append([tag, m_group, mat])
                hl_no = hl_no + 1
        pdf.table(
            tbl,
            row,
            ["type", "group", "form"],
            rc="symbol",
            caption="Atomic SAMB.",
            cpos="c|c|c|c",
            stretch=1.3,
            long=True,
            rmath=True,
            hl=hl[1:],
        )

        # cluster SAMB.
        if not info["molecule"]:
            self.comment(pdf, "cluster SAMB.")
            row = []
            tbl = []
            hl_no = -1
            hl = []
            for c_group, c_lst in samb["info"]["cluster"].items():
                hl.append(hl_no)
                c_group = self.ubar_name(c_group)
                for cmp in c_lst:
                    tag, vec = samb["data"]["cluster"][cmp]
                    tag = "$" + TagMultipole(tag).latex() + "$"
                    no = str(int(cmp.split("_")[1]))
                    no = "\mathbb{Y}_{" + no + "}"
                    vec = "$" + NSArray(vec).latex() + "$"
                    row.append(no)
                    tbl.append([tag, c_group, vec])
                    hl_no = hl_no + 1
            pdf.table(
                tbl,
                row,
                ["type", "cluster", "form"],
                rc="symbol",
                caption="Cluster SAMB.",
                cpos="c|c|c|c",
                stretch=1.3,
                long=True,
                rmath=True,
                hl=hl[1:],
            )

        # uniform SAMB.
        self.comment(pdf, "uniform SAMB.")
        row = []
        tbl = []
        hl_no = -1
        hl = []
        for c_group, u_lst in samb["info"]["uniform"].items():
            hl.append(hl_no)
            c_group = self.ubar_name(c_group)
            for ump in u_lst:
                tag, shape, mat = samb["data"]["uniform"][ump]
                tag = "$" + TagMultipole(tag).latex() + "$"
                no = str(int(ump.split("_")[1]))
                no = "\mathbb{U}_{" + no + "}"
                mat = "$" + convert_to_matrix(shape[0], shape[1], mat).latex() + "$"
                row.append(no)
                tbl.append([tag, c_group, mat])
                hl_no = hl_no + 1
        pdf.table(
            tbl,
            row,
            ["type", "cluster", "form"],
            rc="symbol",
            caption="Uniform SAMB.",
            cpos="c|c|c|c",
            stretch=1.3,
            long=True,
            rmath=True,
            hl=hl[1:],
        )

        if not info["molecule"]:
            # structure SAMB.
            self.comment(pdf, "structure SAMB.")
            row = []
            tbl = []
            hl_no = -1
            hl = []
            for c_group, s_lst in samb["info"]["structure"].items():
                hl.append(hl_no)
                c_group = self.ubar_name(c_group)
                for smp in s_lst:
                    tag, val = samb["data"]["structure"][smp]
                    tag = "$" + TagMultipole(tag).latex() + "$"
                    no = str(int(smp.split("_")[1]))
                    no = "\mathbb{F}_{" + no + "}"
                    val = "$" + NSArray(val).latex() + "$"
                    row.append(no)
                    tbl.append([tag, c_group, val])
                    hl_no = hl_no + 1
            pdf.table(
                tbl,
                row,
                ["type", "cluster", "form"],
                rc="symbol",
                caption="Structure SAMB.",
                cpos="c|c|c|c",
                stretch=1.3,
                long=True,
                rmath=True,
                hl=hl[1:],
            )

    # ==================================================
    def harmonics_info(self, pdf, info, samb):
        self.comment(pdf, "harmonics.")
        group = self._mpm.point_group

        polar = TagList.from_str(TagMultipole, samb["info"]["harmonics"]["Q"])
        axial = TagList.from_str(TagMultipole, samb["info"]["harmonics"]["G"])

        for hset, name in [(polar, "Polar"), (axial, "Axial")]:
            if not hset:
                continue
            row = []
            tbl = []
            hl_no = -1
            hl = []
            old_rank = -1
            for no, h_tag in enumerate(hset):
                head = h_tag.head
                rank = h_tag.rank
                mul = h_tag.mul
                comp = h_tag.comp
                irrep = h_tag.tag_irrep()
                dim = irrep.dim
                if mul < 1:
                    mul = "-"
                if dim == 1:
                    comp = "-"
                harm = group.harmonics[h_tag]
                harm_ltx = harm.latex().replace("h,", "")
                tbl.append([harm_ltx, str(rank), irrep.latex(), mul, comp, harm.expression(v=NSArray.vector3d(head)).latex()])
                row.append(str(no + 1))
                if rank != old_rank:
                    old_rank = rank
                    hl.append(hl_no)
                hl_no = hl_no + 1
            pdf.table(
                tbl,
                row,
                ["symbol", "rank", "irrep.", "mul.", "comp.", "form"],
                "No.",
                caption=f"{name} harmonics.",
                stretch=1.3,
                rmath=True,
                tmath=True,
                long=True,
                hl=hl[1:],
            )

    # ==================================================
    def lattice_info(self, pdf, info, name, data):
        self.comment(pdf, "unit cell.")
        pdf.text(r"\item Unit cell:" + "\n")
        cell = info["cell"]
        a, b, c, alpha, beta, gamma = (
            float(cell["a"]),
            float(cell["b"]),
            float(cell["c"]),
            float(cell["alpha"]),
            float(cell["beta"]),
            float(cell["gamma"]),
        )
        txt = f"\quad $a={a},\,\, b={b},\,\, c={c},\,\, \\alpha={alpha},\,\, \\beta={beta},\,\, \\gamma={gamma}$\n"
        pdf.text(txt)

        pdf.text(r"\item Lattice vectors:" + "\n")
        a1, a2, a3 = NSArray(info["a1"]), NSArray(info["a2"]), NSArray(info["a3"])
        pdf.text(r"\quad $\bm{a}_1=" + a1.evalf().latex() + "$\n")
        pdf.text(r"\quad $\bm{a}_2=" + a2.evalf().latex() + "$\n")
        pdf.text(r"\quad $\bm{a}_3=" + a3.evalf().latex() + "$\n")
        if len(data["plus_set"]) > 1:
            pdf.text(r"\item Plus sets:" + "\n")
            for i in data["plus_set"]:
                pdf.text(r"\quad $+" + NSArray(i).latex() + "$\n")

        self.comment(pdf, "k path.")
        kpath = info["k_path"].replace("Γ", "$\Gamma$")
        tbl = sum([[k.replace("Γ", "$\Gamma$"), "$" + NSArray(v).latex() + "$"] for k, v in info["k_point"].items()], [])
        ncol = min(3, len(info["k_point"]))
        pdf.table(
            tbl,
            [""],
            ["symbol", "position"] * ncol,
            caption="High-symmetry line: " + kpath + ".",
            cpos="c" + "|cc" * ncol,
            stretch=1.3,
            long=True,
        )

    # ==================================================
    def group_info(self, pdf, info, name, data):
        group = self._mpm.group
        so = group.symmetry_operation

        self.comment(pdf, "group information.")
        pdf.text(r"\item Group info.: Generator = " + "$" + ",\,\,".join(so.gen.latex()) + "$" + "\n")

        # conjugacy class.
        irop = [i.latex() for i in so.cc]
        row = []
        tbl = []
        cap = "Conjugacy class"
        if not group.tag.is_point_group():
            cap += " (point-group part)"
        cap += "."
        for ir in irop:
            row.append("$" + ir[0] + "$")
            tbl.append(["$" + "$,\,\, $".join(ir) + "$"])
        pdf.table(
            tbl,
            row,
            ["symmetry operations"],
            rc="rep. SO",
            caption=cap,
            cpos="c|l",
            stretch=1.3,
            long=True,
            hl=True,
        )

        # symmetry operation.
        tbl = []
        for i, op in enumerate(so.full):
            tbl.append([str(i + 1), "$" + op.latex() + "$"])
        tbl = sum(tbl, [])
        pdf.table(
            tbl,
            [""],
            col=["No.", "SO"] * 5,
            caption="Symmetry operations.",
            cpos="c" + "|cc" * 5,
            stretch=1.3,
            long=True,
        )

        # character.
        if not group.tag.is_point_group():
            ch = group.pg.character
        else:
            ch = group.character
        col = ch.symmetry_operation().latex()
        row = ch.irrep_list.latex()
        tbl = [sympy_to_latex(ch.character(i)) for i in ch.irrep_list]
        cap = "Character table"
        if not group.tag.is_point_group():
            cap += " (point-group part)"
        cap += "."
        align = "c|" + "r" * len(col)
        pdf.table(tbl, row, col, caption=cap, rcmath=True, rmath=True, cmath=True, tmath=True, cpos=align, long=True)

        # parity conversion.
        tbl = [i.latex() + "\\,\\,(" + j.latex() + ")" for i, j in ch.parity_conversion().items()]
        pdf.table(tbl, [""], ["$\leftrightarrow$"] * min(len(tbl), 5), caption="Parity conversion.", tmath=True, long=True)

        # symmetric product.
        tbl = []
        for i, ir1 in enumerate(ch.irrep_list):
            t1 = []
            for j, ir2 in enumerate(ch.irrep_list):
                v = sp.latex(ch.symmetric_product_decomposition((ir1, ir2), ret_ex=True)) if j >= i else ""
                t1.append(v)
            tbl.append(t1)
        cp = "c|" + "c" * len(row)
        pdf.table(
            tbl,
            row,
            row,
            rmath=True,
            cmath=True,
            tmath=True,
            cpos=cp,
            caption="Symmetric product, $[\Gamma\otimes\Gamma']_+.$",
            long=True,
        )

        # anti-symmetric product.
        tbl = []
        for ir in ch.irrep_list:
            v = sp.latex(ch.anti_symmetric_product_decomposition(ir, ret_ex=True))
            if v == "0":
                v = "-"
            tbl.append(v)
        pdf.table(
            [tbl],
            [""],
            row,
            rmath=True,
            cmath=True,
            tmath=True,
            caption="Anti-symmetric product, $[\Gamma\otimes\Gamma]_-$.",
            long=True,
        )

        # virtual cluster.
        self.hr(pdf)

        pdf.text("{\n" + r"\scriptsize")

        if info["molecule"]:
            vc = group.virtual_cluster
        else:
            vc = group.pg.virtual_cluster
        sites = sum([[str(no + 1), i.latex()] for no, i in enumerate(vc.site)], [])
        col = ["No.", "position"] * min(4, len(sites) // 2)
        pdf.table(sites, [""], col, caption="Virtual-cluster sites.", tmath=True, stretch=1.7, long=True)

        col = [str(i + 1) for i in range(min(10, len(sites) // 2))]
        row = []
        tbl = []
        for info, basis in vc.items():
            row.append("$" + info.latex().replace("s,", "") + "$")
            basis = NSArray(basis.tolist(), "scalar")
            tbl.append(basis.latex())
        pdf.table(tbl, row, col, "symbol", hl=True, caption="Virtual-cluster basis.", tmath=True, stretch=1.7, long=True)

        pdf.text("}")



__def_dict__ = {
    "subgroup": {
        "triclinic": "cubic",
        "monoclinic": "cubic",
        "orthorhombic": "cubic",
        "tetragonal": "cubic",
        "trigonal": "hexagonal",
        "hexagonal": "hexagonal",
        "cubic": "cubic",
    },  # subgroup.
    "parentgroup": {
        "cubic": ["triclinic", "monoclinic", "orthorhombic", "tetragonal", "cubic"],
        "hexagonal": ["trigonal", "hexagonal"],
    },  # parent group.
    "crystal": ["triclinic", "monoclinic", "orthorhombic", "tetragonal", "trigonal", "hexagonal", "cubic"],  # crystal system.
    "head": ["Q", "G", "T", "M"],  # multipole type.
    "inversion": ["polar", "axial"],  # inversion type.
    "time_reversal": ["electric", "magnetic"],  # time-reversal type.
    "head_i": {"Q": "polar", "T": "polar", "M": "axial", "G": "axial"},  # parity of head.
    "head_t": {"Q": "electric", "T": "magnetic", "M": "magnetic", "G": "electric"},  # time-reversal property of head.
    "it_head": {("polar", "electric"): "Q", ("axial", "electric"): "G", ("polar", "magnetic"): "T", ("axial", "magnetic"): "M"},
    "head_tr": {"Q": "T", "T": "Q", "M": "G", "G": "M"},  # toggle time-reversal property.
    "head_ir": {"Q": "G", "T": "M", "M": "T", "G": "Q"},  # toggle parity.
    "head_harm": {"Q": "Q", "T": "Q", "M": "G", "G": "G"},  # conversion to harmonics head.
    "to_polar": {"Q": "Q", "T": "T", "M": "T", "G": "Q"},  # conversion to polar head.
    "irrep_dim": {"A": 1, "B": 1, "E": 2, "T": 3},  # dimension of irrep. (head letter).
    "response_head": {  # (rank,type) to response tensor head.
        (0, "s"): "C",
        (1, "s"): "C",
        (2, "s"): "S",  # symmetric part.
        (2, "a"): "A",  # anti-symmetric part.
        (3, "s"): "S",  # symmetric part.
        (3, "a"): "A",  # anti-symmetric part.
        (4, "sss"): "S",  # symmetric-symmetric (symmetric) part.
        (4, "ssa"): "Sb",  # symmetric-symmetric (anti-symmetric) part.
        (4, "aas"): "A",  # anti-symmetric-anti-symmetric (symmetric) part.
        (4, "aaa"): "Ab",  # anti-symmetric-anti-symmetric (anti-symmetric) part.
        (4, "sa"): "M",  # symmetric-anti-symmetric part.
        (4, "as"): "Mb",  # anti-symmetric-symmetric part.
    },
}


def is_electric(tag):
    return tag == __def_dict__["time_reversal"][0] or tag in ["Q", "G"]


def is_magnetic(tag):
    return tag == __def_dict__["time_reversal"][1] or tag in ["T", "M"]


def is_polar(tag):
    return tag == __def_dict__["inversion"][0] or tag in ["Q", "T"]


def is_axial(tag):
    return tag == __def_dict__["inversion"][1] or tag in ["G", "M"]



"""
TagResponseTensor manages tag of response tensor.
"""
from dataclasses import dataclass
from multipie.tag.tag import Tag
from multipie.const import __def_dict__


# ==================================================
@dataclass(frozen=True, order=True)
class TagResponseTensor(Tag):
    """
    tag of response tensor.
    """

    head: str = "Q"  #
    rank: int = 0  #
    comp: str = "s"  # rank(0):(s), (1):(s), (2):(s/a), (3):(s/a), (4):(sss/ssa/aas/aaa/sa/as).

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        tensor_head, tag = tag.split("^")
        rank, tag = tag[2:].split(",")
        rank = int(rank)
        head = tag[0]
        comp = ""
        for (r, c), th in __def_dict__["response_head"].items():
            if r == rank and th == tensor_head:
                comp = c
                break
        d = {
            "head": head,
            "rank": rank,
            "comp": comp,
        }
        return d

    # ==================================================
    def __str__(self):
        return TagResponseTensor._tensor_str(self.head, self.rank, self.comp)

    # ==================================================
    def latex(self):
        ss = str(self).replace("Sb", "\\bar{S}").replace("Ab", "\\bar{A}").replace("Mb", "\\bar{M}")
        return ss

    # ==================================================
    @property
    def i_type(self):
        """
        parity type.

        Returns:
            str: parity type, (polar/axial).
        """
        return __def_dict__["head_i"][self.head]

    # ==================================================
    @property
    def t_type(self):
        """
        time-reversal type.

        Returns:
            str: time-reversal type, (electric/magnetic).
        """
        return __def_dict__["head_t"][self.head]

    # ==================================================
    @property
    def tensor_head(self):
        return __def_dict__["response_head"][(self.rank, self.comp)]

    # ==================================================
    @classmethod
    def _tensor_str(cls, head, rank, comp):
        ss = __def_dict__["response_head"][(rank, comp)] + "^{(" + str(rank) + "," + head + ")}"
        return ss

    # ==================================================
    @classmethod
    def create(cls, head, rank, comp):
        """
        create response tensor tag.

        Args:
            head (str): head, (Q/G/T/M).
            rank (int): rank, (0/1/2/3/4).
            comp (str): (0):(s), (1):(s), (2):(s/a), (3):(s/a), (4):(sss/ssa/aas/aaa/sa/as).

        Returns:
            TagResponseTensor: tag.
        """
        return TagResponseTensor(TagResponseTensor._tensor_str(head, rank, comp))



"""
ResponseTensorPG manages physical response tensors upto rank 4.
"""
import sympy as sp
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_list import TagList
from multipie.response_tensor.util.response_tensor_util import simplify_tensor, create_tensor
from multipie.tag.tag_response_tensor import TagResponseTensor
from multipie.const import __def_dict__


# ==================================================
class ResponseTensorPG(dict):  # dict of (response tag, matrix), { TagResponseTensor: NSArray }.
    """
    a set of response tensors upto rank 4.

    Attributes:
        tag (TagGroup): point-group tag.
        definition ({TagResponseTensor:{sympy:sympy}}): definition in terms of multipoles for each component.
    """

    # ==================================================
    def __init__(self, pg_tag):
        """
        initialize the class.

        Args:
            pg_tag (TagGroup or str): point group tag.
        """
        if pg_tag is None:
            self.tag = None
            """point-group tag."""
            self.definition = {}
        else:
            self.tag = TagGroup(str(pg_tag))
            self._initialize()

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def _initialize(self):
        """
        create response tensors.
        """
        # create response tensor data, { tag: [[(tensor symbol, component)]] }.
        d = {}
        for head in __def_dict__["head"]:
            for rank, comp in __def_dict__["response_head"].keys():
                tag = TagResponseTensor.create(head, rank, comp)
                d[tag] = create_tensor(self.tag, tag)

        # create active tensor, { tag: Matrix }.
        mat = {}
        for tag, M in d.items():
            M = simplify_tensor(M)
            zero = sp.zeros(len(M), len(M[0]))
            x = zero.copy()
            for i, lst in enumerate(M):
                for j, (c, m) in enumerate(lst):
                    if m == sp.S(0):
                        x[i, j] = sp.S(0)
                    else:
                        x[i, j] = c
            if not x.equals(zero):
                mat[tag] = x

        # create definition of each element, { tag: {tensor_symbol: definition} }.
        de = {}
        for tag, M1 in mat.items():
            de[tag] = []
            M2 = sp.Matrix([[m for _, m in lst] for lst in d[tag]])
            Cs = []

            for i in range(M1.rows):
                for j in range(M1.cols):
                    m = M1[i, j]
                    if m != sp.S(0):
                        Cs += [i for i in m.atoms(sp.Symbol)]
            Cs = list(set(Cs))
            for m1, m2 in zip(M1, M2):
                c_lst = [i for i in m1.atoms(sp.Symbol)]
                if all([c in Cs for c in c_lst]) and m1 != sp.S(0):
                    de[tag].append((m1, m2))
                    for c in c_lst:
                        Cs.remove(c)

        for k, v in mat.items():
            self[k] = NSArray(v.tolist(), "matrix")
        self.definition = {k: dict(v) for k, v in de.items()}
        """definition in terms of multipoles for each component."""

    # ==================================================
    def select(self, **kwargs):
        """
        select response tensors with given keywords.

        Args:
            kwargs (dict): select conditions for response tensors, (head/rank/comp).

        Returns:
            ResponseTensorPG: selected response tensors.
        """
        c_rt = ResponseTensorPG(None)
        for t in self.key_list().select(**kwargs):
            c_rt[t] = self[t]
            c_rt.definition[t] = self.definition[t]
        c_rt.tag = self.tag

        return c_rt

    # ==================================================
    def _dump(self):
        """
        dump response tensors.
        """
        for tag in self.keys():
            print(f"=== rank {tag.rank} {tag.i_type} {tag.t_type} tensor '{tag.comp}' component ===")
            print(tag.symbol(), "=", self[tag])

            def_mul = self.definition[tag]
            for t, d in def_mul.items():
                print("   ", t, ":=", d)

    # ==================================================
    def _info(self):
        """
        get information of the response tensors.
        """
        s = "* -------------------------------------------------------------------------------------------------------- *" + "\n"
        s += " The symmetry properties of the response tensors 'C' of the 0-th to 4-th orders are summarized" + "\n"
        s += " in terms of the electric, electric toroidal, magnetic, and magnetic toroidal multipoles, Q, G, M, and T." + "\n"
        s += " The basis of the matrices is represented by using the Voigt notation (1=xx, 2=yy, 3=zz, 4=yz, 5=zx, 6=xy)." + "\n"
        s += "* -------------------------------------------------------------------------------------------------------- *" + "\n"
        s += "- rank 0 tensor" + "\n"
        s += "  C = Q_{0}" + "\n"
        s += "    C = (Q_{0})" + "\n"
        s += "- rank 1 tensor" + "\n"
        s += "  C = Q_{1}" + "\n"
        s += "    C = (Q_{x}, Q_{y}, Q_{z})" + "\n"
        s += "- rank 2 tensor" + "\n"
        s += "  C12 = S12 + A12" + "\n"
        s += "    S12 = S21" + "\n"
        s += "    A12 = -A21" + "\n"
        s += "- rank 3 tensor" + "\n"
        s += "  C123 = S12;3 + A12;3" + "\n"
        s += "    S12;3 = S21;3" + "\n"
        s += "    A12;3 = -A21;3" + "\n"
        s += "- rank 4 tensor" + "\n"
        s += "  C1234 = S12;34 + Sb12;34 + A12;34 + Ab12;34 + M12;34 + Mb12;34" + "\n"
        s += "    S12;34 = S21;34 = S12;43 = S34;12" + "\n"
        s += "    Sb12;34 = Sb21;34 = Sb12;43 = -Sb34;12" + "\n"
        s += "    A12;34 = -A21;34 = -A12;43 = A34;12" + "\n"
        s += "    Ab12;34 = -Ab21;34 = -Ab12;43 = -Ab34;12" + "\n"
        s += "    M12;34 = M21;34 = -M12;43" + "\n"
        s += "    Mb12;34 = -Mb21;34 = Mb12;43" + "\n"

        print(s)

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagResponseTensor(tag))
        else:
            return self.get(tag)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())



"""
This file provides utility functions for calculation of structure multipole basis set.
"""
import sympy as sp
import numpy as np

from gcoreutils.nsarray import NSArray


# ==================================================
def cs_list(s, real=False):
    """
    list of c### and s### contained in the given equation.

    Args:
        s (str/Symbol): string object to be converted to symbol.
        real (bool): real symbol ? If real=True, the object is recognized as a real number.

    Returns:
        list: [c###/s###]
    """
    s = sp.expand(sp.sympify(str(s)))
    cs_lst = []
    n = 1
    while s != 0:
        cn, sn = sp.Symbol(f"c{n:03d}"), sp.Symbol(f"s{n:03d}")
        c = sp.diff(s, cn)
        s = sp.limit(s, cn, 0)
        if c != 0:
            cs_lst.append(cn)
        c = sp.diff(s, sn)
        s = sp.limit(s, sn, 0)
        if c != 0:
            cs_lst.append(sn)
        n += 1

    if real:
        cs_lst = [sp.Symbol(str(csn), real=True) for csn in cs_lst]

    return cs_lst


# ==================================================
def to_symbol(s, real=False):
    """
    convert string to symbol.

    Args:
        s (str): string object to be converted to symbol.
        real (bool): real symbol ? If real=True, the object is recognized as a real number.

    Returns:
        Symbol: symbol.
    """
    cs_lst = cs_list(s)
    s = sp.expand(sp.sympify(str(s)))
    d = {sp.Symbol(str(csn), real=real): s.coeff(csn) for csn in cs_lst}
    return sp.expand(sum([c * csn for csn, c in d.items()]))


# ==================================================
def inner_product(f1, f2):
    """
    inner product of two structure factors.

    Args:
        f1 (str/Symbol): F1.
        f2 (str/Symbol): F2.

    Returns:
        NSArray: inner product of F1 and F2.

    Notes:
        - F1 and F2 are polynomials in terms of c### and s### (### are given by lst) only without const. terms.
    """
    f1 = str(to_symbol(f1))
    f2 = str(to_symbol(f2))

    ip_str = {}
    for csi in cs_list(f1):
        for csj in cs_list(f2):
            if csi != csj:
                ip_str[str(csi * csj)] = "(0)"
            else:
                ip_str[str(csi * csj)] = "(1/2)"

    prod = str((NSArray(f1) * NSArray(f2)).expand())
    for k, v in ip_str.items():
        prod = prod.replace(k, v)

    prod = sp.expand(sp.sympify(prod))

    return prod


# ==================================================
def orthogonalize_fk(v, nmax=None):
    """
    orthogonalize list of function of k by Gram-Schmidt orthogonalization method.

    Args:
        v (NSArray): list of function of k to be orthogonalized.
        nmax (int, optional): max. number of nonzero basis.

    Returns:
        NSArray: orthogonalized list of function of k.
    """
    ev = []

    cnt = 0
    for v0 in v:
        s = sp.S(0)
        for evi in ev:
            s += inner_product(evi, v0) * evi

        e = v0 - s
        e = sp.expand(e)
        d = sp.sqrt(inner_product(e, e))
        if d != 0:
            e = e / d
            e = sp.expand(e)
            cnt += 1
        else:
            e = sp.S(0)
        ev.append(e)
        if nmax is not None and cnt == nmax:
            break

    idx = np.array([i for i, b in enumerate(ev) if not b == sp.S(0)])

    return ev, idx


# ==================================================
def decompose_fk(fk, k_samb_set):
    """
    decompose k function into linear combination of structure multipoles.

    Args:
        fk (str/Symbol): function of k.
        k_samb_set (list): [ ("kmp_#", "formfactor") ].

    Returns:
        dict: dictionary of expansion coefficient {"smp_#": coeff}.
    """
    fk = sp.expand(to_symbol(fk))
    if fk == sp.S(0):
        return {}

    sm_set = {kmp_i: sp.expand(fk_) for kmp_i, fk_ in k_samb_set}

    coeffs = {}
    for kmp_i, fk_ in sm_set.items():
        c = inner_product(fk, fk_)
        if c != 0:
            coeffs[kmp_i] = c

    return coeffs


def test():
    c1, s1, c2, s2 = sp.symbols("c001 s001 c002 s002")
    eq = sp.expand(sp.sqrt(2) * (c1 - sp.I * s1) + sp.sqrt(3) * (c2 + sp.I * s2))
    print(f"eq = {eq}")

    cs_lst = cs_list(eq)
    print(cs_lst)

    eq_ = sp.conjugate(to_symbol(str(eq), real=True))
    print(f"eq_ = {eq_}")

    f1 = sp.expand(sp.sqrt(2) * (c1 - s1))
    f2 = sp.expand(sp.sqrt(3) * (c1 + s1))
    print(f"inner_product(f1, f1) = {inner_product(f1, f1)}")
    print(f"inner_product(f1, f2) = {inner_product(f1, f2)}")
    print(f"inner_product(f2, f2) = {inner_product(f2, f2)}")

    v = [f1, f2]
    v = orthogonalize_fk(v)[0]
    print(f"orthogonalize_fk = {v}")

    k_samb_set = [("kmp_1", v[0]), ("kmp_2", v[1])]
    print(f"decompose_fk = {decompose_fk(c1, k_samb_set)}")


# test()



"""
This file provides utility functions for combined multipole set.
"""
from itertools import product

from gcoreutils.nsarray import NSArray

from multipie.tag.tag_list import TagList
from multipie.const import is_electric


# ==================================================
def create_z_samb(cg, hs, tag1_list, tag2_list, toroidal_priority=False, **kwargs):
    """
    initialize combined multipoles.

    Args:
        cg (clebsch_gordan_pg): Clebsh-Gordan coefficients.
        hs (HarmonicsPG): the class to manage harmonics.
        tag1_list (TagList): multipole/harmonics tag list.
        tag2_list (TagList): multipole/harmonics tag list.
        toroidal_priority (bool): create toroidal multipoles (G,T) in priority? else prioritize conventional multipoles (Q,M).
        kwargs (dict, optional): select conditions for multipoles,
                                 keywords in ["head", "rank", "irrep", "mul", "comp", "s", "k"].
    Returns:
        dict: {TagMultipole: [(coefficient, TagMultipole(atomic), TagMultipole(site/bond)] }.
    """
    key_list = ["head", "rank", "irrep", "mul", "comp", "s", "k"]
    for key in kwargs.keys():
        if key not in key_list:
            raise Exception(f"{key} cannot be specified in kwargs.")

    kwargs = {k: [v] if type(v) != list else v for k, v in kwargs.items()}
    keys, values = list(kwargs.keys()), list(kwargs.values())
    idx_pairs = product(*[range(len(v_list)) for v_list in values])
    kwargs_list = [{keys[i]: values[i][j] for i, j in enumerate(idx_pair)} for idx_pair in idx_pairs]

    head_list = ["G", "Q", "T", "M"] if toroidal_priority else ["Q", "G", "M", "T"]
    Ztag_list = [tag.replace(m_type="") for tag in hs.key_list()]
    Ztag_list += [tag.replace(head="M") for tag in TagList(Ztag_list).select(head="G")]
    Ztag_list += [tag.replace(head="T") for tag in TagList(Ztag_list).select(head="Q")]

    Ztag_list = [tag for kwargs in kwargs_list for tag in TagList(Ztag_list).select(**kwargs)]
    Ztag_list = sorted(set(Ztag_list), key=Ztag_list.index)
    Ztag_list = [tag for l in range(12) for Z in head_list for tag in Ztag_list if tag.rank == l and tag.head == Z]
    Ztag_list = TagList(Ztag_list)

    is_inversion = True if "g" in Ztag_list[0].irrep or "u" in Ztag_list[0].irrep else False
    irrep_comp_list = [(tag.irrep, tag.comp) for tag in Ztag_list] if is_inversion else [tag.irrep for tag in Ztag_list]
    irrep_comp_list = sorted(set(irrep_comp_list), key=irrep_comp_list.index)

    head_s_k_l_list1 = [(tag1.head, tag1.s, tag1.k, tag1.l) for tag1 in tag1_list]
    head_s_k_l_list1 = sorted(set(head_s_k_l_list1), key=head_s_k_l_list1.index)
    head_l_list2 = [(tag2.head, tag2.l) for tag2 in tag2_list]
    head_l_list2 = sorted(set(head_l_list2), key=head_l_list2.index)

    z_samb = {}
    for X, s, k, l1 in head_s_k_l_list1:
        tag1_list_ = tag1_list.select(head=X, s=s, k=k, rank=l1)
        for Y, l2 in head_l_list2:
            tag2_list_ = tag2_list.select(head=Y, rank=l2)
            tag1_tag2_list = [(tag1, tag2) for tag1 in tag1_list_ for tag2 in tag2_list_]
            dim = len(tag1_list_) * len(tag2_list_)
            min_l, max_l = abs(l1 - l2), l1 + l2
            t1 = +1 if is_electric(X) else -1
            t2 = +1 if is_electric(Y) else -1
            t_type = "electric" if t1 * t2 == 1 else "magnetic"

            for irrep_comp in irrep_comp_list:
                if is_inversion:
                    irrep, comp = irrep_comp
                    Ztag_list_ = Ztag_list.select(irrep=irrep, comp=comp)
                else:
                    irrep = irrep_comp
                    Ztag_list_ = Ztag_list.select(irrep=irrep)

                Ztag_list_ = [tag for tag in Ztag_list_ if tag.t_type == t_type and min_l <= tag.l and tag.l <= max_l]

                tags, vecs = [], []
                for tag in Ztag_list_:
                    lst = []
                    for tag1, tag2 in tag1_tag2_list:
                        c = cg.cg(tag1, tag2, tag)
                        lst.append(c)

                    if lst != [0] * dim:
                        vecs.append(NSArray(lst, style="vector"))
                        tags.append(tag.replace(s=s, k=k))

                # orthogonalize
                if vecs != []:
                    vecs = NSArray(vecs, style="vector")
                    vecs, idx = NSArray.orthogonalize(vecs, dim)
                    vecs = vecs[idx]
                    tags = [tags[i] for i in idx]

                    for tag, vec in zip(tags, vecs):
                        lst = []
                        for i, c in enumerate(vec):
                            if c != 0:
                                tag1, tag2 = tag1_tag2_list[i]
                                lst.append((c, tag1, tag2))
                        _, tag1, tag2 = lst[0]
                        z_samb[(tag, tag1, tag2)] = lst

    return z_samb
